# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ducktape.mark import parametrize, matrix, ignore, defaults
from ducktape.mark.mark_expander import MarkedFunctionExpander
from ducktape.mark.resource import cluster
from ducktape.cluster.cluster_spec import ClusterSpec

import pytest


class CheckClusterUseAnnotation(object):

    def check_basic_usage_arbitrary_metadata(self):
        cluster_use_metadata = {
            'a': 2,
            'b': 'hi',
            'num_nodes': 300
        }

        @cluster(**cluster_use_metadata)
        def function():
            return "hi"
        assert hasattr(function, "marks")

        test_context_list = MarkedFunctionExpander(function=function).expand()
        assert len(test_context_list) == 1
        assert test_context_list[0].cluster_use_metadata == cluster_use_metadata

    def check_basic_usage_cluster_spec(self):
        num_nodes = 200

        @cluster(cluster_spec=ClusterSpec.simple_linux(num_nodes))
        def function():
            return "hi"
        assert hasattr(function, "marks")

        test_context_list = MarkedFunctionExpander(function=function).expand()
        assert len(test_context_list) == 1
        assert len(test_context_list[0].expected_cluster_spec.nodes.os_to_nodes) == 1
        assert len(test_context_list[0].expected_cluster_spec.nodes.os_to_nodes.get('linux')) == num_nodes

    def check_basic_usage_num_nodes(self):
        num_nodes = 200

        @cluster(num_nodes=num_nodes)
        def function():
            return "hi"
        assert hasattr(function, "marks")

        test_context_list = MarkedFunctionExpander(function=function).expand()
        assert len(test_context_list) == 1
        assert test_context_list[0].expected_num_nodes == num_nodes

    def check_with_parametrize(self):
        num_nodes = 200

        @cluster(num_nodes=num_nodes)
        @parametrize(x=1)
        def f(x, y=2, z=3):
            return x, y, z

        test_context_list = MarkedFunctionExpander(function=f).expand()
        assert len(test_context_list) == 1
        assert test_context_list[0].expected_num_nodes == num_nodes

    def check_beneath_parametrize(self):
        """ Annotations such as cluster, which add metadata, but do not create new tests, add the metadata to
        all test cases physically below the annotation.

        In the case of a parametrized test, when @cluster has no parametrize annotations below it,
        there are not any test cases to which it will apply, so none of the resulting tests should have
        associated cluster size metadata.
        """
        num_nodes = 200

        @parametrize(x=1)
        @cluster(num_nodes=num_nodes)
        def f(x, y=2, z=3):
            return x, y, z

        with pytest.raises(AssertionError):
            MarkedFunctionExpander(function=f).expand()

    def check_with_nodes_default_parametrize_matrix(self):
        default_num_nodes = 5
        parametrize_num_nodes = 3
        matrix_num_nodes = 7

        @cluster(num_nodes=default_num_nodes)
        @defaults(y=[5])
        @parametrize(x=100, y=200)
        @cluster(num_nodes=parametrize_num_nodes)
        @parametrize(x=10, y=20)
        @parametrize(x=30, y=40)
        @cluster(num_nodes=matrix_num_nodes)
        @matrix(x=[1, 2])
        def f(x, y):
            return x, y

        test_context_list = MarkedFunctionExpander(function=f).expand()
        assert len(test_context_list) == 5
        # {'x': 1, 'y': 5} -> using matrix with matrix_num_nodes
        assert test_context_list[0].expected_num_nodes == matrix_num_nodes
        # {'x': 2, 'y': 5} -> using matrix with matrix_num_nodes
        assert test_context_list[1].expected_num_nodes == matrix_num_nodes
        # {'x': 30, 'y': 40} -> using parametrize with cluster parametrize_num_nodes
        assert test_context_list[2].expected_num_nodes == parametrize_num_nodes
        # {'x': 10, 'y': 20} -> using parametrize with cluster parametrize_num_nodes
        assert test_context_list[3].expected_num_nodes == parametrize_num_nodes
        # {'x': 100, 'y': 200} -> using parametrize with default_num_nodes
        assert test_context_list[4].expected_num_nodes == default_num_nodes

    def check_no_override(self):
        """ cluster use metadata should apply to all test cases physically below it, except for those which already
        have cluster use metadata.
        """

        num_nodes_1 = 200
        num_nodes_2 = 42

        # num_nodes_2 should *not* override num_nodes_1
        @cluster(num_nodes=num_nodes_2)
        @cluster(num_nodes=num_nodes_1)
        def f(x, y=2, z=3):
            return x, y, z

        test_context_list = MarkedFunctionExpander(function=f).expand()
        assert len(test_context_list) == 1
        assert test_context_list[0].expected_num_nodes == num_nodes_1

    def check_parametrized_with_multiple_cluster_annotations(self):
        num_nodes_1 = 10
        num_nodes_2 = 20

        # num_nodes_1 should *not* override num_nodes_2
        @cluster(num_nodes=num_nodes_1)
        @parametrize(x=1)
        @parametrize(x=1.5)
        @cluster(num_nodes=num_nodes_2)
        @parametrize(x=2)
        def f(x, y=2, z=3):
            return x, y, z

        test_context_list = MarkedFunctionExpander(function=f).expand()
        assert len(test_context_list) == 3
        assert test_context_list[0].expected_num_nodes == num_nodes_1
        assert test_context_list[1].expected_num_nodes == num_nodes_1
        assert test_context_list[2].expected_num_nodes == num_nodes_2

    def check_matrix_with_multiple_cluster_annotations(self):
        num_nodes_1 = 10
        num_nodes_2 = 20

        # num_nodes_1 should *not* override num_nodes_2
        @cluster(num_nodes=num_nodes_1)
        @matrix(x=[1])
        @matrix(x=[1.5, 1.6], y=[-15, -16])
        @cluster(num_nodes=num_nodes_2)
        @matrix(x=[2])
        def f(x, y=2, z=3):
            return x, y, z

        test_context_list = MarkedFunctionExpander(function=f).expand()
        assert len(test_context_list) == 6
        assert test_context_list[0].expected_num_nodes == num_nodes_1
        assert test_context_list[1].expected_num_nodes == num_nodes_1
        assert test_context_list[2].expected_num_nodes == num_nodes_1
        assert test_context_list[3].expected_num_nodes == num_nodes_1
        assert test_context_list[4].expected_num_nodes == num_nodes_1
        assert test_context_list[5].expected_num_nodes == num_nodes_2

    def check_with_ignore(self):
        num_nodes = 200

        @cluster(num_nodes=num_nodes)
        @ignore
        def function():
            return "hi"
        assert hasattr(function, "marks")

        test_context_list = MarkedFunctionExpander(function=function).expand()
        assert len(test_context_list) == 1
        assert test_context_list[0].expected_num_nodes == num_nodes

        # order shouldn't matter here
        @ignore
        @cluster(num_nodes=num_nodes)
        def function():
            return "hi"
        assert hasattr(function, "marks")

        test_context_list = MarkedFunctionExpander(function=function).expand()
        assert len(test_context_list) == 1
        assert test_context_list[0].expected_num_nodes == num_nodes



# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster_spec import ClusterSpec
from ducktape.command_line.defaults import ConsoleDefaults
from ducktape.template import TemplateRenderer
from ducktape.errors import TimeoutError

import os
import shutil
import tempfile
import time


class ServiceIdFactory:
    def generate_service_id(self, service):
        return "{service_name}-{service_number}-{service_id}".format(
            service_name=service.__class__.__name__,
            service_number=service._order,
            service_id=id(service)
        )


class MultiRunServiceIdFactory:
    def __init__(self, run_number=1):
        self.run_number = run_number

    def generate_service_id(self, service):
        return "{run_number}-{service_name}-{service_number}-{service_id}".format(
            run_number=self.run_number,
            service_name=service.__class__.__name__,
            service_number=service._order,
            service_id=id(service)
        )


service_id_factory = ServiceIdFactory()


class Service(TemplateRenderer):
    """Service classes know how to deploy a service onto a set of nodes and then clean up after themselves.

    They request the necessary resources from the cluster,
    configure each node, and bring up/tear down the service.

    They also expose
    information about the service so that other services or test scripts can
    easily be configured to work with them. Finally, they may be able to collect
    and check logs/output from the service, which can be helpful in writing tests
    or benchmarks.

    Services should generally be written to support an arbitrary number of nodes,
    even if instances are independent of each other. They should be able to assume
    that there won't be resource conflicts: the cluster tests are being run on
    should be large enough to use one instance per service instance.
    """

    # Provides a mechanism for locating and collecting log files produced by the service on its nodes.
    # logs is a dict with entries that look like log_name: {"path": log_path, "collect_default": boolean}
    #
    # For example, zookeeper service might have self.logs like this:
    # self.logs = {
    #    "zk_log": {"path": "/mnt/zk.log",
    #               "collect_default": True}
    # }
    logs = {}

    def __init__(self, context, num_nodes=None, cluster_spec=None, *args, **kwargs):
        """
        Initialize the Service.

        Note: only one of (num_nodes, cluster_spec) may be set.

        :param context:         An object which has at minimum 'cluster' and 'logger' attributes. In tests, this
                                is always a TestContext object.
        :param num_nodes:       An integer representing the number of Linux nodes to allocate.
        :param cluster_spec:    A ClusterSpec object representing the minimum cluster specification needed.
        """
        super(Service, self).__init__(*args, **kwargs)
        # Keep track of significant events in the lifetime of this service
        self._init_time = time.time()
        self._start_time = -1
        self._start_duration_seconds = -1
        self._stop_time = -1
        self._stop_duration_seconds = -1
        self._clean_time = -1

        self._initialized = False
        self.service_id_factory = service_id_factory
        self.cluster_spec = Service.setup_cluster_spec(num_nodes=num_nodes, cluster_spec=cluster_spec)
        self.context = context

        self.nodes = []
        self.skip_nodes_allocation = kwargs.get("skip_nodes_allocation", False)
        if not self.skip_nodes_allocation:
            self.allocate_nodes()

        # Keep track of which nodes nodes were allocated to this service, even after nodes are freed
        # Note: only keep references to representations of the nodes, not the actual node objects themselves
        self._nodes_formerly_allocated = [str(node.account) for node in self.nodes]

        # Every time a service instance is created, it registers itself with its
        # context object. This makes it possible for external mechanisms to clean up
        # after the service if something goes wrong.
        #
        # Note: Allocate nodes *before* registering self with the service registry
        self.context.services.append(self)

        # Each service instance has its own local scratch directory on the test driver
        self._local_scratch_dir = None
        self._initialized = True

    @staticmethod
    def setup_cluster_spec(num_nodes=None, cluster_spec=None):
        if num_nodes is None:
            if cluster_spec is None:
                raise RuntimeError("You must set either num_nodes or cluster_spec.")
            else:
                return cluster_spec
        else:
            if cluster_spec is not None:
                raise RuntimeError("You must set only one of (num_nodes, cluster_spec)")
            return ClusterSpec.simple_linux(num_nodes)

    def __repr__(self):
        return "<%s: %s>" % (self.who_am_i(), "num_nodes: %d, nodes: %s" %
                             (len(self.nodes), [n.account.hostname for n in self.nodes]))

    @property
    def num_nodes(self):
        return len(self.nodes)

    @property
    def local_scratch_dir(self):
        """This local scratch directory is created/destroyed on the test driver before/after each test is run."""
        if not self._local_scratch_dir:
            self._local_scratch_dir = tempfile.mkdtemp()
        return self._local_scratch_dir

    @property
    def service_id(self):
        """Human-readable identifier (almost certainly) unique within a test run."""
        return self.service_id_factory.generate_service_id(self)

    @property
    def _order(self):
        """Index of this service instance with respect to other services of the same type registered with self.context.
        When used with a test_context, this lets the user know

        Example::

            suppose the services registered with the same context looks like
                context.services == [Zookeeper, Kafka, Zookeeper, Kafka, MirrorMaker]
            then:
                context.services[0]._order == 0  # "0th" Zookeeper instance
                context.services[2]._order == 0  # "0th" Kafka instance
                context.services[1]._order == 1  # "1st" Zookeeper instance
                context.services[3]._order == 1  # "1st" Kafka instance
                context.services[4]._order == 0  # "0th" MirrorMaker instance

        """
        if hasattr(self.context, "services"):
            same_services = [id(s) for s in self.context.services if type(s) == type(self)]

            if self not in self.context.services and not self._initialized:
                # It's possible that _order will be invoked in the constructor *before* self has been registered with
                # the service registry (aka self.context.services).
                return len(same_services)

            # Note: index raises ValueError if the item is not in the list
            index = same_services.index(id(self))
            return index
        else:
            return 0

    @property
    def logger(self):
        """The logger instance for this service."""
        return self.context.logger

    @property
    def cluster(self):
        """The cluster object from which this service instance gets its nodes."""
        return self.context.cluster

    @property
    def allocated(self):
        """Return True iff nodes have been allocated to this service instance."""
        return len(self.nodes) > 0

    def who_am_i(self, node=None):
        """Human-readable identifier useful for log messages."""
        if node is None:
            return self.service_id
        else:
            return "%s node %d on %s" % (self.service_id, self.idx(node), node.account.hostname)

    def allocate_nodes(self):
        """Request resources from the cluster."""
        if self.allocated:
            raise Exception("Requesting nodes for a service that has already been allocated nodes.")

        self.logger.debug("Requesting nodes from the cluster: %s" % self.cluster_spec)

        try:
            self.nodes = self.cluster.alloc(self.cluster_spec)
        except RuntimeError as e:
            msg = str(e)
            if hasattr(self.context, "services"):
                msg += " Currently registered services: " + str(self.context.services)
            raise RuntimeError(msg)

        for idx, node in enumerate(self.nodes, 1):
            # Remote accounts utilities should log where this service logs
            if node.account._logger is not None:
                # This log message help test-writer identify which test and/or service didn't clean up after itself
                node.account.logger.critical(ConsoleDefaults.BAD_TEST_MESSAGE)
                raise RuntimeError(
                    "logger was not None on service start. There may be a concurrency issue, "
                    "or some service which isn't properly cleaning up after itself. "
                    "Service: %s, node.account: %s" % (self.__class__.__name__, str(node.account)))
            node.account.logger = self.logger

        self.logger.debug("Successfully allocated %d nodes to %s" % (len(self.nodes), self.who_am_i()))

    def start(self, **kwargs):
        """Start the service on all nodes."""
        self.logger.info("%s: starting service" % self.who_am_i())
        if self._start_time < 0:
            # Set self._start_time only the first time self.start is invoked
            self._start_time = time.time()

        self.logger.debug(self.who_am_i() + ": killing processes and attempting to clean up before starting")
        for node in self.nodes:
            # Added precaution - kill running processes, clean persistent files (if 'clean'=False flag passed,
            # skip cleaning), try/except for each step, since each of these steps may fail if there
            # are no processes to kill or no files to remove

            try:
                self.stop_node(node)
            except Exception:
                pass

            try:
                if kwargs.get('clean', True):
                    self.clean_node(node)
                else:
                    self.logger.debug("%s: skip cleaning node" % self.who_am_i(node))
            except Exception:
                pass

        for node in self.nodes:
            self.logger.debug("%s: starting node" % self.who_am_i(node))
            self.start_node(node, **kwargs)

        if self._start_duration_seconds < 0:
            self._start_duration_seconds = time.time() - self._start_time

    def start_node(self, node, **kwargs):
        """Start service process(es) on the given node."""
        pass

    def wait(self, timeout_sec=600):
        """Wait for the service to finish.
        This only makes sense for tasks with a fixed amount of work to do. For services that generate
        output, it is only guaranteed to be available after this call returns.
        """
        unfinished_nodes = []
        start = time.time()
        end = start + timeout_sec
        for node in self.nodes:
            now = time.time()
            node_name = self.who_am_i(node)
            if end > now:
                self.logger.debug("%s: waiting for node", node_name)
                if not self.wait_node(node, end - now):
                    unfinished_nodes.append(node_name)
            else:
                unfinished_nodes.append(node_name)

        if unfinished_nodes:
            raise TimeoutError("Timed out waiting %s seconds for service nodes to finish. " % str(timeout_sec)
                               + "These nodes are still alive: " + str(unfinished_nodes))

    def wait_node(self, node, timeout_sec=None):
        """Wait for the service on the given node to finish.
        Return True if the node finished shutdown, False otherwise.
        """
        pass

    def stop(self, **kwargs):
        """Stop service processes on each node in this service.
        Subclasses must override stop_node.
        """
        self._stop_time = time.time()  # The last time stop is invoked
        self.logger.info("%s: stopping service" % self.who_am_i())
        for node in self.nodes:
            self.logger.info("%s: stopping node" % self.who_am_i(node))
            self.stop_node(node, **kwargs)

        self._stop_duration_seconds = time.time() - self._stop_time

    def stop_node(self, node, **kwargs):
        """Halt service process(es) on this node."""
        pass

    def clean(self, **kwargs):
        """Clean up persistent state on each node - e.g. logs, config files etc.
        Subclasses must override clean_node.
        """
        self._clean_time = time.time()
        self.logger.info("%s: cleaning service" % self.who_am_i())
        for node in self.nodes:
            self.logger.info("%s: cleaning node" % self.who_am_i(node))
            self.clean_node(node, **kwargs)

    def clean_node(self, node, **kwargs):
        """Clean up persistent state on this node - e.g. service logs, configuration files etc."""
        self.logger.warn("%s: clean_node has not been overriden. "
                         "This may be fine if the service leaves no persistent state."
                         % self.who_am_i())

    def free(self):
        """Free each node. This 'deallocates' the nodes so the cluster can assign them to other services."""
        while self.nodes:
            node = self.nodes.pop()
            self.logger.info("%s: freeing node" % self.who_am_i(node))
            node.account.logger = None
            self.cluster.free(node)

    def run(self):
        """Helper that executes run(), wait(), and stop() in sequence."""
        self.start()
        self.wait()
        self.stop()

    def get_node(self, idx):
        """ids presented externally are indexed from 1, so we provide a helper method to avoid confusion."""
        return self.nodes[idx - 1]

    def idx(self, node):
        """Return id of the given node. Return -1 if node does not belong to this service.

        idx identifies the node within this service instance (not globally).
        """
        for idx, n in enumerate(self.nodes, 1):
            if self.get_node(idx) == node:
                return idx
        return -1

    def close(self):
        """Release resources."""
        # Remove local scratch directory
        if self._local_scratch_dir and os.path.exists(self._local_scratch_dir):
            shutil.rmtree(self._local_scratch_dir)

    @staticmethod
    def run_parallel(*args):
        """Helper to run a set of services in parallel. This is useful if you want
           multiple services of different types to run concurrently, e.g. a
           producer + consumer pair.
        """
        for svc in args:
            svc.start()
        for svc in args:
            svc.wait()
        for svc in args:
            svc.stop()

    def to_json(self):
        return {
            "cls_name": self.__class__.__name__,
            "module_name": self.__module__,

            "lifecycle": {
                "init_time": self._init_time,
                "start_time": self._start_time,
                "start_duration_seconds": self._start_duration_seconds,
                "stop_time": self._stop_time,
                "stop_duration_seconds": self._stop_duration_seconds,
                "clean_time": self._clean_time
            },
            "service_id": self.service_id,
            "nodes": self._nodes_formerly_allocated
        }



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.command_line.main import get_user_defined_globals
from ducktape.command_line.main import setup_results_directory
from ducktape.command_line.main import update_latest_symlink

import json
import os
import os.path
import pickle
import pytest
import tempfile


class CheckSetupResultsDirectory(object):
    def setup_method(self, _):
        self.results_root = tempfile.mkdtemp()
        self.results_dir = os.path.join(self.results_root, "results_directory")
        self.latest_symlink = os.path.join(self.results_root, "latest")

    def validate_directories(self):
        """Validate existence of results directory and correct symlink"""
        assert os.path.exists(self.results_dir)
        assert os.path.exists(self.latest_symlink)
        assert os.path.islink(self.latest_symlink)

        # check symlink points to correct location
        assert os.path.realpath(self.latest_symlink) == os.path.realpath(self.results_dir)

    def check_creation(self):
        """Check results and symlink from scratch"""
        assert not os.path.exists(self.results_dir)
        setup_results_directory(self.results_dir)
        update_latest_symlink(self.results_root, self.results_dir)
        self.validate_directories()

    def check_symlink(self):
        """Check "latest" symlink behavior"""

        # if symlink already exists
        old_results = os.path.join(self.results_root, "OLD")
        os.mkdir(old_results)
        os.symlink(old_results, self.latest_symlink)
        assert os.path.islink(self.latest_symlink) and os.path.exists(self.latest_symlink)

        setup_results_directory(self.results_dir)
        update_latest_symlink(self.results_root, self.results_dir)
        self.validate_directories()

        # Try again if symlink exists and points to nothing
        os.rmdir(self.results_dir)
        assert os.path.islink(self.latest_symlink) and not os.path.exists(self.latest_symlink)
        setup_results_directory(self.results_dir)
        update_latest_symlink(self.results_root, self.results_dir)
        self.validate_directories()


globals_json = """
{
    "x": 200
}
"""

invalid_globals_json = """
{
    can't parse this!: ?right?
}
"""

valid_json_not_dict = """
[
    {
        "x": 200,
        "y": 300
    }
]
"""


class CheckUserDefinedGlobals(object):
    """Tests for the helper method which parses in user defined globals option"""

    def check_immutable(self):
        """Expect the user defined dict object to be immutable."""
        global_dict = get_user_defined_globals(globals_json)

        with pytest.raises(NotImplementedError):
            global_dict["x"] = -1

        with pytest.raises(NotImplementedError):
            global_dict["y"] = 3

    def check_pickleable(self):
        """Expect the user defined dict object to be pickleable"""
        globals_dict = get_user_defined_globals(globals_json)

        assert globals_dict  # Need to test non-empty dict, to ensure py3 compatibility
        assert pickle.loads(pickle.dumps(globals_dict)) == globals_dict

    def check_parseable_json_string(self):
        """Check if globals_json is parseable as JSON, we get back a dictionary view of parsed JSON."""
        globals_dict = get_user_defined_globals(globals_json)
        assert globals_dict == json.loads(globals_json)

    def check_unparseable(self):
        """If globals string is not a path to a file, and not parseable as JSON we want to raise a ValueError
        """
        with pytest.raises(ValueError):
            get_user_defined_globals(invalid_globals_json)

    def check_parse_from_file(self):
        """Validate that, given a filename of a file containing valid JSON, we correctly parse the file contents."""
        _, fname = tempfile.mkstemp()
        try:
            with open(fname, "w") as fh:
                fh.write(globals_json)

            global_dict = get_user_defined_globals(fname)
            assert global_dict == json.loads(globals_json)
            assert global_dict["x"] == 200
        finally:
            os.remove(fname)

    def check_bad_parse_from_file(self):
        """Validate behavior when given file containing invalid JSON"""
        _, fname = tempfile.mkstemp()
        try:
            with open(fname, "w") as fh:
                # Write invalid JSON
                fh.write(invalid_globals_json)

            with pytest.raises(ValueError):
                get_user_defined_globals(fname)

        finally:
            os.remove(fname)

    def check_non_dict(self):
        """Valid JSON which does not parse as a dict should raise a ValueError"""

        # Should be able to parse this as JSON
        json.loads(valid_json_not_dict)

        with pytest.raises(ValueError):
            get_user_defined_globals(valid_json_not_dict)

    def check_non_dict_from_file(self):
        """Validate behavior when given file containing valid JSON which does not parse as a dict"""
        _, fname = tempfile.mkstemp()
        try:
            with open(fname, "w") as fh:
                # Write valid JSON which does not parse as a dict
                fh.write(valid_json_not_dict)

            with pytest.raises(ValueError, match=fname):
                get_user_defined_globals(fname)

        finally:
            os.remove(fname)



# Copyright 2021 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.localhost import LocalhostCluster
from tests.ducktape_mock import test_context, session_context

import logging
import pytest
from ducktape.tests.runner_client import Sender
from ducktape.tests.runner import Receiver
from ducktape.tests.event import ClientEventFactory, EventResponseFactory
from ducktape.errors import TimeoutError

import multiprocessing as mp
import os


class CheckSenderReceiver(object):
    def ready_response(self, client_id, port):
        sender_event_factory = ClientEventFactory("test_1", 0, client_id)
        sender = Sender(server_host='localhost', server_port=port,
                        message_supplier=sender_event_factory, logger=logging)
        sender.send(sender_event_factory.ready())

    def check_simple_messaging(self):
        s_context = session_context()
        cluster = LocalhostCluster(num_nodes=1000)
        t_context = test_context(s_context, cluster)

        client_id = "test-runner-{}-{}".format(os.getpid(), id(self))
        receiver_response_factory = EventResponseFactory()

        receiver = Receiver(5556, 5656)
        receiver.start()
        port = receiver.port

        try:
            p = mp.Process(target=self.ready_response, args=(client_id, port))
            p.start()

            event = receiver.recv(timeout=10000)
            assert event["event_type"] == ClientEventFactory.READY
            logging.info('replying to client')
            receiver.send(receiver_response_factory.ready(event, s_context, t_context, cluster))
        finally:
            p.join()

    def check_timeout(self):
        client_id = "test-runner-{}-{}".format(os.getpid(), id(self))

        receiver = Receiver(5556, 5656)
        receiver.start()
        port = receiver.port

        try:
            p = mp.Process(target=self.ready_response, args=(client_id, port))
            p.start()
            with pytest.raises(TimeoutError):
                receiver.recv(timeout=0)
        finally:
            p.join()



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import signal
import time
import traceback
import zmq

from six import iteritems
from ducktape.services.service import MultiRunServiceIdFactory, service_id_factory
from ducktape.services.service_registry import ServiceRegistry

from ducktape.tests.event import ClientEventFactory
from ducktape.tests.loader import TestLoader
from ducktape.tests.serde import SerDe
from ducktape.tests.status import FLAKY
from ducktape.tests.test import test_logger, TestContext

from ducktape.tests.result import TestResult, IGNORE, PASS, FAIL
from ducktape.utils.local_filesystem_utils import mkdir_p


def run_client(*args, **kwargs):
    client = RunnerClient(*args, **kwargs)
    client.run()


class RunnerClient(object):
    """Run a single test"""

    def __init__(self, server_hostname, server_port, test_id,
                 test_index, logger_name, log_dir, debug, fail_bad_cluster_utilization, deflake_num):
        signal.signal(signal.SIGTERM, self._sigterm_handler)  # register a SIGTERM handler

        self.serde = SerDe()
        self.logger = test_logger(logger_name, log_dir, debug)
        self.runner_port = server_port

        self.fail_bad_cluster_utilization = fail_bad_cluster_utilization
        self.test_id = test_id
        self.test_index = test_index
        self.id = "test-runner-%d-%d" % (os.getpid(), id(self))
        self.message = ClientEventFactory(self.test_id, self.test_index, self.id)
        self.sender = Sender(server_hostname, str(self.runner_port), self.message, self.logger)

        ready_reply = self.sender.send(self.message.ready())
        self.session_context = ready_reply["session_context"]
        self.test_metadata = ready_reply["test_metadata"]
        self.cluster = ready_reply["cluster"]

        self.deflake_num = deflake_num

        # Wait to instantiate the test object until running the test
        self.test = None
        self.test_context = None
        self.all_services = None

    def send(self, event):
        return self.sender.send(event)

    def _sigterm_handler(self, signum, frame):
        """Translate SIGTERM to SIGINT on this process

        python will treat SIGINT as a Keyboard exception. Exception handling does the rest.
        """
        os.kill(os.getpid(), signal.SIGINT)

    def _collect_test_context(self, directory, file_name, cls_name, method_name, injected_args):
        loader = TestLoader(self.session_context, self.logger, injected_args=injected_args, cluster=self.cluster)
        # TODO: deal with this in a more graceful fashion.
        #       In an unlikely even that discover either raises the exception or fails to find exactly one test
        #       we should probably continue trying other tests rather than killing this process
        loaded_context_list = loader.discover(directory, file_name, cls_name, method_name)

        assert len(loaded_context_list) == 1
        test_context = loaded_context_list[0]
        test_context.cluster = self.cluster
        return test_context

    def run(self):
        self.log(logging.INFO, "Loading test %s" % str(self.test_metadata))
        self.test_context = self._collect_test_context(**self.test_metadata)
        self.test_context.test_index = self.test_index

        self.send(self.message.running())
        if self.test_context.ignore:
            # Skip running this test, but keep track of the fact that we ignored it
            result = TestResult(self.test_context,
                                self.test_index,
                                self.session_context,
                                test_status=IGNORE,
                                start_time=time.time(),
                                stop_time=time.time())
            result.report()
            # Tell the server we are finished
            self.send(self.message.finished(result=result))
            return

        start_time = -1
        stop_time = -1
        test_status = FAIL
        summary = []
        data = None
        self.all_services = ServiceRegistry()

        num_runs = 0

        try:
            while test_status == FAIL and num_runs < self.deflake_num:
                num_runs += 1
                self.log(logging.INFO, "on run {}/{}".format(num_runs, self.deflake_num))
                start_time = time.time()
                test_status, summary, data = self._do_run(num_runs)

                if test_status == PASS and num_runs > 1:
                    test_status = FLAKY

                msg = str(test_status.to_json())
                if summary:
                    msg += ": {}".format(summary)
                if num_runs != self.deflake_num:
                    msg += "\n" + "~" * max(len(line) for line in summary.split('\n'))

                self.log(logging.INFO, msg)

        finally:
            stop_time = time.time()

            test_status, summary = self._check_cluster_utilization(test_status, summary)

            if num_runs > 1:
                # for reporting purposes report all services
                self.test_context.services = self.all_services
            # for flaky tests, we report the start and end time of the successful run, and not the whole run period
            result = TestResult(
                self.test_context,
                self.test_index,
                self.session_context,
                test_status,
                summary,
                data,
                start_time,
                stop_time)

            self.log(logging.INFO, "Data: %s" % str(result.data))

            result.report()
            # Tell the server we are finished
            self._do_safely(lambda: self.send(self.message.finished(result=result)),
                            "Problem sending FINISHED message for " + str(self.test_metadata) + ":\n")
            # Release test_context resources only after creating the result and finishing logging activity
            # The Sender object uses the same logger, so we postpone closing until after the finished message is sent
            self.test_context.close()
            self.all_services = None
            self.test_context = None
            self.test = None

    def _do_run(self, num_runs):
        test_status = FAIL
        summary = []
        data = None
        sid_factory = MultiRunServiceIdFactory(num_runs) if self.deflake_num > 1 else service_id_factory
        try:
            # Results from this test, as well as logs will be dumped here
            mkdir_p(TestContext.results_dir(self.test_context, self.test_index))
            # Instantiate test
            self.test = self.test_context.cls(self.test_context)
            # Check if there are enough nodes
            self._check_min_cluster_spec()
            # Run the test unit

            self.setup_test()

            data = self.run_test()

            test_status = PASS

        except BaseException as e:
            # mark the test as failed before doing anything else
            test_status = FAIL
            err_trace = self._exc_msg(e)
            summary.append(err_trace)

        finally:
            for service in self.test_context.services:
                service.service_id_factory = sid_factory
                self.all_services.append(service)

            self.teardown_test(teardown_services=not self.session_context.no_teardown, test_status=test_status)

            if hasattr(self.test_context, "services"):
                service_errors = self.test_context.services.errors()
                if service_errors:
                    summary.extend(["\n\n", service_errors])

            # free nodes
            if self.test:
                self.log(logging.DEBUG, "Freeing nodes...")
                self._do_safely(self.test.free_nodes, "Error freeing nodes:")
            return test_status, "".join(summary), data

    def _check_min_cluster_spec(self):
        self.log(logging.DEBUG, "Checking if there are enough nodes...")
        min_cluster_spec = self.test.min_cluster_spec()
        os_to_num_nodes = {}
        for node_spec in min_cluster_spec:
            if not os_to_num_nodes.get(node_spec.operating_system):
                os_to_num_nodes[node_spec.operating_system] = 1
            else:
                os_to_num_nodes[node_spec.operating_system] = os_to_num_nodes[node_spec.operating_system] + 1
        for (operating_system, node_count) in iteritems(os_to_num_nodes):
            num_avail = len(list(self.cluster.all().nodes.elements(operating_system=operating_system)))
            if node_count > num_avail:
                raise RuntimeError(
                    "There are not enough nodes available in the cluster to run this test. "
                    "Cluster size for %s: %d, Need at least: %d. Services currently registered: %s" %
                    (operating_system, num_avail, node_count, self.test_context.services))

    def _check_cluster_utilization(self, result, summary):
        """Checks if the number of nodes used by a test is less than the number of
        nodes requested by the test. If this is the case and we wish to fail
        on bad cluster utilization, the result value is failed. Will also print
        a warning if the test passes and the node utilization doesn't match.
        """
        max_used = self.cluster.max_used()
        total = len(self.cluster.all())
        if max_used < total:
            message = "Test requested %d nodes, used only %d" % (total, max_used)
            if self.fail_bad_cluster_utilization:
                # only check node utilization on test pass
                if result == PASS or result == FLAKY:
                    self.log(logging.INFO, "FAIL: " + message)

                result = FAIL
                summary += message
            else:
                self.log(logging.WARN, message)
        return result, summary

    def setup_test(self):
        """start services etc"""
        self.log(logging.INFO, "Setting up...")
        self.test.setup()

    def run_test(self):
        """Run the test!

        We expect test_context.function to be a function or unbound method which takes an
        instantiated test object as its argument.
        """
        self.log(logging.INFO, "Running...")
        return self.test_context.function(self.test)

    def _exc_msg(self, e):
        return repr(e) + "\n" + traceback.format_exc(limit=16)

    def _do_safely(self, action, err_msg):
        try:
            action()
        except BaseException as e:
            self.log(logging.WARN, err_msg + " " + self._exc_msg(e))

    def teardown_test(self, teardown_services=True, test_status=None):
        """teardown method which stops services, gathers log data, removes persistent state, and releases cluster nodes.

        Catch all exceptions so that every step in the teardown process is tried, but signal that the test runner
        should stop if a keyboard interrupt is caught.
        """
        self.log(logging.INFO, "Tearing down...")
        if not self.test:
            self.log(logging.WARN, "%s failed to instantiate" % self.test_id)
            self.test_context.close()
            return

        services = self.test_context.services

        if teardown_services:
            self._do_safely(self.test.teardown, "Error running teardown method:")
            # stop services
            self._do_safely(services.stop_all, "Error stopping services:")

        # always collect service logs whether or not we tear down
        # logs are typically removed during "clean" phase, so collect logs before cleaning
        self.log(logging.DEBUG, "Copying logs from services...")
        self._do_safely(lambda: self.test.copy_service_logs(test_status), "Error copying service logs:")

        # clean up stray processes and persistent state
        if teardown_services:
            self.log(logging.DEBUG, "Cleaning up services...")
            self._do_safely(services.clean_all, "Error cleaning services:")

    def log(self, log_level, msg, *args, **kwargs):
        """Log to the service log and the test log of the current test."""

        if self.test_context is None:
            msg = "%s: %s" % (self.__class__.__name__, str(msg))
            self.logger.log(log_level, msg, *args, **kwargs)
        else:
            msg = "%s: %s: %s" % (self.__class__.__name__, self.test_context.test_name, str(msg))
            self.logger.log(log_level, msg, *args, **kwargs)

        self.send(self.message.log(msg, level=log_level))


class Sender(object):
    REQUEST_TIMEOUT_MS = 3000
    NUM_RETRIES = 5

    def __init__(self, server_host, server_port, message_supplier, logger):
        self.serde = SerDe()
        self.server_endpoint = "tcp://%s:%s" % (str(server_host), str(server_port))
        self.zmq_context = zmq.Context()
        self.socket = None
        self.poller = zmq.Poller()

        self.message_supplier = message_supplier
        self.logger = logger

        self._init_socket()

    def _init_socket(self):
        self.socket = self.zmq_context.socket(zmq.REQ)
        self.socket.connect(self.server_endpoint)
        self.poller.register(self.socket, zmq.POLLIN)

    def send(self, event, blocking=True):

        retries_left = Sender.NUM_RETRIES

        while retries_left > 0:
            serialized_event = self.serde.serialize(event)
            self.socket.send(serialized_event)
            retries_left -= 1
            waiting_for_reply = True

            while waiting_for_reply:
                sockets = dict(self.poller.poll(Sender.REQUEST_TIMEOUT_MS))

                if sockets.get(self.socket) == zmq.POLLIN:
                    reply = self.socket.recv()
                    if reply:
                        return self.serde.deserialize(reply)
                    else:
                        # send another request...
                        break
                else:
                    self.close()
                    self._init_socket()
                    waiting_for_reply = False
                # Ensure each message we attempt to send has a unique id
                # This copy constructor gives us a duplicate with a new message id
                event = self.message_supplier.copy(event)

        raise RuntimeError("Unable to receive response from driver")

    def close(self):
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.close()
        self.poller.unregister(self.socket)



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle


class SerDe(object):
    def serialize(self, obj):
        if hasattr(obj, 'serialize'):
            obj.serialize()
        else:
            return pickle.dumps(obj)

    def deserialize(self, bytes_obj, obj_cls=None):
        if obj_cls and hasattr(obj_cls, 'deserialize'):
            return obj_cls.deserialize(bytes_obj)
        else:
            return pickle.loads(bytes_obj)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from contextlib import contextmanager
import logging
import os
import re
import shutil
import sys
import tempfile

from ducktape.cluster.cluster_spec import ClusterSpec
from ducktape.tests.loggermaker import LoggerMaker, close_logger
from ducktape.utils.local_filesystem_utils import mkdir_p
from ducktape.command_line.defaults import ConsoleDefaults
from ducktape.services.service_registry import ServiceRegistry
from ducktape.template import TemplateRenderer
from ducktape.mark.resource import CLUSTER_SPEC_KEYWORD, CLUSTER_SIZE_KEYWORD
from ducktape.tests.status import FAIL


class Test(TemplateRenderer):
    """Base class for tests.
    """

    def __init__(self, test_context, *args, **kwargs):
        """
        :type test_context: ducktape.tests.test.TestContext
        """
        super(Test, self).__init__(*args, **kwargs)
        self.test_context = test_context

    @property
    def cluster(self):
        return self.test_context.cluster

    @property
    def logger(self):
        return self.test_context.logger

    def min_cluster_spec(self):
        """
        Returns a specification for the minimal cluster we need to run this test.

        This method replaces the deprecated min_cluster_size.  Unlike min_cluster_size, it can handle
        non-Linux operating systems.

        In general, most Tests don't need to override this method.  The default implementation
        seen here works well in most cases.  However, the default implementation only takes into account
        the services that exist at the time of the call.  You may need to override this method if you add
        new services during the course of your test.

        :return:            A ClusterSpec object.
        """
        try:
            # If the Test overrode the deprecated min_cluster_size method, we will use that.
            num_linux_nodes = self.min_cluster_size()
            return ClusterSpec.simple_linux(num_linux_nodes)
        except NotImplementedError:
            # Otherwise, ask the service registry what kind of cluster spec we need for currently
            # extant services.
            return self.test_context.services.min_cluster_spec()

    def min_cluster_size(self):
        """
        Returns the number of linux nodes which this test needs.

        THIS METHOD IS DEPRECATED, and provided only for backwards compatibility.
        Please implement min_cluster_spec instead.

        :return:            An integer.
        """
        raise NotImplementedError

    def setup(self):
        """Override this for custom setup logic."""

        # for backward compatibility
        self.setUp()

    def teardown(self):
        """Override this for custom teardown logic."""

        # for backward compatibility
        self.tearDown()

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def free_nodes(self):
        try:
            self.test_context.services.free_all()
        except BaseException as e:
            if isinstance(e, KeyboardInterrupt):
                raise

    def compress_service_logs(self, node, service, node_logs):
        """Compress logs on a node corresponding to the given service.

        :param node: The node on which to compress the given logs
        :param service: The service to which the node belongs
        :param node_logs: Paths to logs (or log directories) which will be compressed
        :return: a list of paths to compressed logs.

        """
        compressed_logs = []
        for nlog in node_logs:
            try:
                node.account.ssh(_compress_cmd(nlog))
                if nlog.endswith(os.path.sep):
                    nlog = nlog[:-len(os.path.sep)]
                nlog += ".tgz"
                compressed_logs.append(nlog)

            except Exception as e:
                self.test_context.logger.warn(
                    "Error compressing log %s: service %s: %s" % (nlog, service, str(e))
                )

        return compressed_logs

    def copy_service_logs(self, test_status):
        """
        Copy logs from service nodes to the results directory.

        If the test passed, only the default set will be collected. If the the test failed, all logs will be collected.
        """
        for service in self.test_context.services:
            if not hasattr(service, 'logs') or len(service.logs) == 0:
                self.test_context.logger.debug("Won't collect service logs from %s - no logs to collect." %
                                               service.service_id)
                continue

            log_dirs = service.logs
            for node in service.nodes:
                # Gather locations of logs to collect
                node_logs = []
                for log_name in log_dirs.keys():
                    if test_status == FAIL or self.should_collect_log(log_name, service):
                        node_logs.append(log_dirs[log_name]["path"])

                self.test_context.logger.debug("Preparing to copy logs from %s: %s" %
                                               (node.account.hostname, node_logs))

                if self.test_context.session_context.compress:
                    self.test_context.logger.debug("Compressing logs...")
                    node_logs = self.compress_service_logs(node, service, node_logs)

                if len(node_logs) > 0:
                    # Create directory into which service logs will be copied
                    dest = os.path.join(
                        TestContext.results_dir(self.test_context, self.test_context.test_index),
                        service.service_id, node.account.hostname)
                    if not os.path.isdir(dest):
                        mkdir_p(dest)

                    # Try to copy the service logs
                    self.test_context.logger.debug("Copying logs...")
                    try:
                        for log in node_logs:
                            node.account.copy_from(log, dest)
                    except Exception as e:
                        self.test_context.logger.warn(
                            "Error copying log %(log_name)s from %(source)s to %(dest)s. \
                            service %(service)s: %(message)s" %
                            {'log_name': log_name,
                             'source': log_dirs[log_name],
                             'dest': dest,
                             'service': service,
                             'message': e})

    def mark_for_collect(self, service, log_name=None):
        if log_name is None:
            # Mark every log for collection
            for log_name in service.logs:
                self.test_context.log_collect[(log_name, service)] = True
        else:
            self.test_context.log_collect[(log_name, service)] = True

    def mark_no_collect(self, service, log_name=None):
        self.test_context.log_collect[(log_name, service)] = False

    def should_collect_log(self, log_name, service):
        key = (log_name, service)
        default = service.logs[log_name]["collect_default"]
        val = self.test_context.log_collect.get(key, default)
        return val


def _compress_cmd(log_path):
    """Return bash command which compresses the given path to a tarball."""
    compres_cmd = 'cd "$(dirname %s)" && ' % log_path
    compres_cmd += 'f="$(basename %s)" && ' % log_path
    compres_cmd += 'if [ -e "$f" ]; then tar czf "$f.tgz" "$f"; fi && '
    compres_cmd += 'rm -rf %s' % log_path

    return compres_cmd


def _escape_pathname(s):
    """Remove fishy characters, replace most with dots"""
    # Remove all whitespace completely
    s = re.sub(r"\s+", "", s)

    # Replace bad characters with dots
    blacklist = r"[^\.\-=_\w\d]+"
    s = re.sub(blacklist, ".", s)

    # Multiple dots -> single dot (and no leading or trailing dot)
    s = re.sub(r"[\.]+", ".", s)
    return re.sub(r"^\.|\.$", "", s)


def test_logger(logger_name, log_dir, debug):
    """Helper method for getting a test logger object

    Note that if this method is called multiple times with the same ``logger_name``, it returns the same logger object.
    Note also, that for a fixed ``logger_name``, configuration occurs only the first time this function is called.
    """
    return TestLoggerMaker(logger_name, log_dir, debug).logger


class TestLoggerMaker(LoggerMaker):
    def __init__(self, logger_name, log_dir, debug):
        super(TestLoggerMaker, self).__init__(logger_name)
        self.log_dir = log_dir
        self.debug = debug

    def configure_logger(self):
        """Set up the logger to log to stdout and files.
        This creates a directory and a few files as a side-effect.
        """
        if self.configured:
            return

        self._logger.setLevel(logging.DEBUG)
        mkdir_p(self.log_dir)

        # Create info and debug level handlers to pipe to log files
        info_fh = logging.FileHandler(os.path.join(self.log_dir, "test_log.info"))
        debug_fh = logging.FileHandler(os.path.join(self.log_dir, "test_log.debug"))

        info_fh.setLevel(logging.INFO)
        debug_fh.setLevel(logging.DEBUG)

        formatter = logging.Formatter(ConsoleDefaults.TEST_LOG_FORMATTER)
        info_fh.setFormatter(formatter)
        debug_fh.setFormatter(formatter)

        self._logger.addHandler(info_fh)
        self._logger.addHandler(debug_fh)

        ch = logging.StreamHandler(sys.stdout)
        ch.setFormatter(formatter)
        if self.debug:
            # If debug flag is set, pipe debug logs to stdout
            ch.setLevel(logging.DEBUG)
        else:
            # default - pipe warning level logging to stdout
            ch.setLevel(logging.WARNING)
        self._logger.addHandler(ch)


class TestContext(object):
    """Wrapper class for state variables needed to properly run a single 'test unit'."""

    def __init__(self, **kwargs):
        """
        :param session_context:
        :param cluster: the cluster object which will be used by this test
        :param module: name of the module containing the test class/method
        :param cls: class object containing the test method
        :param function: the test method
        :param file: file containing this module
        :param injected_args: a dict containing keyword args which will be passed to the test method
        :param cluster_use_metadata: dict containing information about how this test will use cluster resources
        """

        self.session_context = kwargs.get("session_context")
        self.cluster = kwargs.get("cluster")
        self.module = kwargs.get("module")
        self.test_suite_name = kwargs.get("test_suite_name")

        if kwargs.get("file") is not None:
            self.file = os.path.abspath(kwargs.get("file"))
        else:
            self.file = None
        self.cls = kwargs.get("cls")
        self.function = kwargs.get("function")
        self.injected_args = kwargs.get("injected_args")
        self.ignore = kwargs.get("ignore", False)

        # cluster_use_metadata is a dict containing information about how this test will use cluster resources
        self.cluster_use_metadata = copy.copy(kwargs.get("cluster_use_metadata", {}))

        self.services = ServiceRegistry()
        self.test_index = None

        # dict for toggling service log collection on/off
        self.log_collect = {}

        self._logger = None
        self._local_scratch_dir = None

    def __repr__(self):
        return \
            "<module=%s, cls=%s, function=%s, injected_args=%s, file=%s, ignore=%s, " \
            "cluster_size=%s, cluster_spec=%s>" % \
            (self.module, self.cls_name, self.function_name, str(self.injected_args), str(self.file),
             str(self.ignore), str(self.expected_num_nodes), str(self.expected_cluster_spec))

    def copy(self, **kwargs):
        """Construct a new TestContext object from another TestContext object
        Note that this is not a true copy, since a fresh ServiceRegistry instance will be created.
        """
        ctx_copy = TestContext(**self.__dict__)
        ctx_copy.__dict__.update(**kwargs)

        return ctx_copy

    @property
    def local_scratch_dir(self):
        """This local scratch directory is created/destroyed on the test driver before/after each test is run."""
        if not self._local_scratch_dir:
            self._local_scratch_dir = tempfile.mkdtemp()
        return self._local_scratch_dir

    @property
    def test_metadata(self):
        return {
            "directory": os.path.dirname(self.file),
            "file_name": os.path.basename(self.file),
            "cls_name": self.cls_name,
            "method_name": self.function_name,
            "injected_args": self.injected_args
        }

    @staticmethod
    def logger_name(test_context, test_index):
        if test_index is None:
            return test_context.test_id
        else:
            return "%s-%s" % (test_context.test_id, str(test_index))

    @staticmethod
    def results_dir(test_context, test_index):
        d = test_context.session_context.results_dir

        if test_context.cls is not None:
            d = os.path.join(d, test_context.cls.__name__)
        if test_context.function is not None:
            d = os.path.join(d, test_context.function.__name__)
        if test_context.injected_args is not None:
            d = os.path.join(d, test_context.injected_args_name)
        if test_index is not None:
            d = os.path.join(d, str(test_index))

        return d

    @property
    def expected_num_nodes(self):
        """
        How many nodes of any type we expect this test to consume when run.

        :return:            an integer number of nodes.
        """
        return self.expected_cluster_spec.size()

    @property
    def expected_cluster_spec(self):
        """
        The cluster spec we expect this test to consume when run.

        :return:            A ClusterSpec object.
        """
        cluster_spec = self.cluster_use_metadata.get(CLUSTER_SPEC_KEYWORD)
        cluster_size = self.cluster_use_metadata.get(CLUSTER_SIZE_KEYWORD)
        if cluster_spec is not None:
            return cluster_spec
        elif cluster_size is not None:
            return ClusterSpec.simple_linux(cluster_size)
        elif self.cluster is None:
            return ClusterSpec.empty()
        else:
            return self.cluster.all()

    @property
    def globals(self):
        return self.session_context.globals

    @property
    def module_name(self):
        return "" if self.module is None else self.module

    @property
    def cls_name(self):
        return "" if self.cls is None else self.cls.__name__

    @property
    def function_name(self):
        return "" if self.function is None else self.function.__name__

    @property
    def description(self):
        """Description of the test, needed in particular for reporting.
        If the function has a docstring, return that, otherwise return the class docstring or "".
        """
        if self.function.__doc__:
            return self.function.__doc__
        elif self.cls.__doc__ is not None:
            return self.cls.__doc__
        else:
            return ""

    @property
    def injected_args_name(self):
        if self.injected_args is None:
            return ""
        else:
            params = ".".join(["%s=%s" % (k, self.injected_args[k]) for k in self.injected_args])
            return _escape_pathname(params)

    @property
    def test_id(self):
        return self.test_name

    @property
    def test_name(self):
        """
        The fully-qualified name of the test. This is similar to test_id, but does not include the session ID. It
        includes the module, class, and method name.
        """
        name_components = [self.module_name,
                           self.cls_name,
                           self.function_name,
                           self.injected_args_name]

        return ".".join(filter(lambda x: x is not None and len(x) > 0, name_components))

    @property
    def logger(self):
        if self._logger is None:
            self._logger = test_logger(
                TestContext.logger_name(self, self.test_index),
                TestContext.results_dir(self, self.test_index),
                self.session_context.debug)
        return self._logger

    def close(self):
        """Release resources, etc."""
        if hasattr(self, "services"):
            for service in self.services:
                service.close()
            # Remove reference to services. This is important to prevent potential memory leaks if users write services
            # which themselves have references to large memory-intensive objects
            del self.services

        # Remove local scratch directory
        if self._local_scratch_dir and os.path.exists(self._local_scratch_dir):
            shutil.rmtree(self._local_scratch_dir)

        # Release file handles held by logger
        if self._logger:
            close_logger(self._logger)


@contextmanager
def in_dir(path):
    """ Changes working directory to given path. On exit, restore to original working directory. """
    cwd = os.getcwd()

    try:
        os.chdir(path)
        yield

    finally:
        os.chdir(cwd)


@contextmanager
def in_temp_dir():
    """ Creates a temporary directory as the working directory. On exit, it is removed. """
    with _new_temp_dir() as tmpdir:
        with in_dir(tmpdir):
            yield tmpdir


@contextmanager
def _new_temp_dir():
    """ Create a temporary directory that is removed automatically """
    tmpdir = tempfile.mkdtemp()

    try:
        yield tmpdir

    finally:
        shutil.rmtree(tmpdir)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ducktape.cluster.cluster_spec import ClusterSpec, WINDOWS, LINUX, NodeSpec
from ducktape.services.service import Service
from ducktape.tests.test import Test
from ducktape.errors import TimeoutError
from ducktape.mark.resource import cluster

import os
import pytest
import random
import shutil
from six import iteritems
import tempfile
from threading import Thread
import time
import logging

from ducktape.utils.util import wait_until


def generate_tempdir_name():
    """Use this ad-hoc function instead of the tempfile module since we're creating and removing
    this directory with ssh commands.
    """
    return "/tmp/" + "t" + str(int(time.time()))


class RemoteAccountTestService(Service):
    """Simple service that allocates one node for performing tests of RemoteAccount functionality"""

    def __init__(self, context):
        super(RemoteAccountTestService, self).__init__(context, num_nodes=1)
        self.temp_dir = generate_tempdir_name()
        self.logs = {
            "my_log": {
                "path": self.log_file,
                "collect_default": True
            },
            "non_existent_log": {
                "path": os.path.join(self.temp_dir, "absent.log"),
                "collect_default": True
            }
        }

    @property
    def log_file(self):
        return os.path.join(self.temp_dir, "test.log")

    def start_node(self, node):
        node.account.ssh("mkdir -p " + self.temp_dir)
        node.account.ssh("touch " + self.log_file)

    def stop_node(self, node):
        pass

    def clean_node(self, node):
        node.account.ssh("rm -rf " + self.temp_dir)

    def write_to_log(self, msg):
        self.nodes[0].account.ssh("echo -e -n " + repr(msg) + " >> " + self.log_file)


class GenericService(Service):
    """Service which doesn't do anything - just a group of nodes, each of which has a scratch directory."""

    def __init__(self, context, num_nodes):
        super(GenericService, self).__init__(context, num_nodes)
        self.worker_scratch_dir = "scratch"
        for node in self.nodes:
            node.account.mkdirs(self.worker_scratch_dir)

    def stop_node(self, node):
        # noop
        pass

    def clean_node(self, node):
        node.account.remove(self.worker_scratch_dir, allow_fail=True)


class UnderUtilizedTest(Test):

    def setup(self):
        self.service = GenericService(self.test_context, 1)

    @cluster(num_nodes=3)
    def under_utilized_test(self):
        # setup() creates a service instance, which calls alloc() for one node
        assert self.test_context.cluster.max_used() == 1
        assert len(self.test_context.cluster.used()) == 1

        self.another_service = GenericService(self.test_context, 1)
        assert len(self.test_context.cluster.used()) == 2
        assert self.test_context.cluster.max_used() == 2

        self.service.stop()
        self.service.free()
        assert len(self.test_context.cluster.used()) == 1
        assert self.test_context.cluster.max_used() == 2


class FileSystemTest(Test):
    """
    Note that in an attempt to isolate the file system methods, validation should be done with ssh/shell commands.
    """

    def setup(self):
        self.service = GenericService(self.test_context, 1)
        self.node = self.service.nodes[0]
        self.scratch_dir = self.service.worker_scratch_dir

    @cluster(num_nodes=1)
    def create_file_test(self):
        expected_contents = "hello world"
        fname = "myfile.txt"
        fpath = "%s/%s" % (self.scratch_dir, fname)

        self.node.account.create_file(fpath, expected_contents)

        # validate existence and contents
        self.node.account.ssh("test -f %s" % fpath)
        contents = "\n".join([line for line in self.node.account.ssh_capture("cat %s" % fpath)])
        assert contents == expected_contents

        # TODO also check absolute path

    @cluster(num_nodes=1)
    def mkdir_test(self):
        dirname = "%s/mydir" % self.scratch_dir
        self.node.account.mkdir(dirname)

        # TODO - important!! check mode
        self.node.account.ssh("test -d %s" % dirname, allow_fail=False)

        # mkdir should not succeed if the base directories do not already exist
        dirname = "%s/a/b/c/d" % self.scratch_dir
        with pytest.raises(IOError):
            self.node.account.mkdir(dirname)

        # TODO also check absolute path

    @cluster(num_nodes=1)
    def mkdirs_nested_test(self):
        dirname = "%s/a/b/c/d" % self.scratch_dir

        # TODO important!! check mode
        self.node.account.mkdirs(dirname)
        self.node.account.ssh("test -d %s" % dirname, allow_fail=False)

        # TODO also check absolute path

    @cluster(num_nodes=1)
    def open_test(self):
        """Try opening, writing, reading a file."""
        fname = "%s/myfile.txt" % self.scratch_dir
        expected_contents = b"hello world\nhooray!"
        with self.node.account.open(fname, "w") as f:
            f.write(expected_contents)

        with self.node.account.open(fname, "r") as f:
            contents = f.read()
        assert contents == expected_contents

        # Now try opening in append mode
        append = b"hithere"
        expected_contents = expected_contents + append
        with self.node.account.open(fname, "a") as f:
            f.write(append)

        with self.node.account.open(fname, "r") as f:
            contents = f.read()

        assert contents == expected_contents

    @cluster(num_nodes=1)
    def exists_file_test(self):
        """
        Create various kinds of files and symlinks, verifying that exists works as expected.

        Note that because
        """

        # create file, test existence with relative and absolute path
        self.node.account.ssh("touch %s/hi" % self.scratch_dir)
        assert self.node.account.exists("%s/hi" % self.scratch_dir)
        # TODO abspath

        # create symlink, test existence with relative and absolute path
        self.node.account.ssh("ln -s %s/hi %s/hi-link" % (self.scratch_dir, self.scratch_dir))
        assert self.node.account.exists("%s/hi-link" % self.scratch_dir)
        # TODO abspath

    def exists_dir_test(self):
        # check bad path doesn't exist
        assert not self.node.account.exists("a/b/c/d")

        # create dir, test existence with relative and absolute path
        dpath = "%s/mydir" % self.scratch_dir
        self.node.account.ssh("mkdir %s" % dpath)
        assert self.node.account.exists(dpath)
        # TODO abspath

        # create symlink, test existence with relative and absolute path
        self.node.account.ssh("ln -s %s %s/mydir-link" % (dpath, self.scratch_dir))
        assert self.node.account.exists("%s/mydir-link" % self.scratch_dir)
        # # TODO abspath

    def remove_test(self):
        """Test functionality of remove method"""
        # remove a non-empty directory
        dpath = "%s/mydir" % self.scratch_dir
        self.node.account.ssh("mkdir %s" % dpath)
        self.node.account.ssh("touch %s/hi.txt" % dpath)
        self.node.account.ssh("test -d %s" % dpath)
        self.node.account.remove(dpath)
        self.node.account.ssh("test ! -d %s" % dpath)

        # remove a file
        fpath = "%s/hello.txt" % self.scratch_dir
        self.node.account.ssh("echo 'hello world' > %s" % fpath)
        self.node.account.remove(fpath)

        # remove non-existent path
        with pytest.raises(RuntimeError):
            self.node.account.remove("a/b/c/d")

        # remove non-existent path with allow_fail = True should be ok
        self.node.account.remove("a/b/c/d", allow_fail=True)


# Representation of a somewhat arbitrary directory structure for testing copy functionality
# A key which has a string as its value represents a file
# A key which has a dict as its value represents a subdirectory
DIR_STRUCTURE = {
    "d00": {
        "another_file": b"1\n2\n3\n4\ncats and dogs",
        "d10": {
            "fasdf": b"lasdf;asfd\nahoppoqnbasnb"
        },
        "d11": {
            "f65": b"afasdfsafdsadf"
        }
    },
    "a_file": b"hello world!"
}


def make_dir_structure(base_dir, dir_structure, node=None):
    """Make a file tree starting at base_dir with structure specified by dir_structure.

    if node is None, make the structure locally, else make it on the given node
    """
    for k, v in iteritems(dir_structure):
        if isinstance(v, dict):
            # it's a subdirectory
            subdir_name = k
            subdir_path = os.path.join(base_dir, subdir_name)
            subdir_structure = v

            if node:
                node.account.mkdir(subdir_path)
            else:
                os.mkdir(subdir_path)

            make_dir_structure(subdir_path, subdir_structure, node)
        else:
            # it's a file
            file_name = k
            file_path = os.path.join(base_dir, file_name)
            file_contents = v

            if node:
                with node.account.open(file_path, "wb") as f:
                    f.write(file_contents)
            else:
                with open(file_path, "wb") as f:
                    f.write(file_contents)


def verify_dir_structure(base_dir, dir_structure, node=None):
    """Verify locally or on the given node whether the file subtree at base_dir matches dir_structure."""
    for k, v in iteritems(dir_structure):
        if isinstance(v, dict):
            # it's a subdirectory
            subdir_name = k
            subdir_path = os.path.join(base_dir, subdir_name)
            subdir_structure = v

            if node:
                assert node.account.isdir(subdir_path)
            else:
                assert os.path.isdir(subdir_path)

            verify_dir_structure(subdir_path, subdir_structure, node)
        else:
            # it's a file
            file_name = k
            file_path = os.path.join(base_dir, file_name)
            expected_file_contents = v

            if node:
                with node.account.open(file_path, "r") as f:
                    contents = f.read()
            else:
                with open(file_path, "rb") as f:
                    contents = f.read()
            assert expected_file_contents == contents, contents


class CopyToAndFroTest(Test):
    """These tests check copy_to, and copy_from functionality."""

    def setup(self):
        self.service = GenericService(self.test_context, 1)
        self.node = self.service.nodes[0]
        self.remote_scratch_dir = self.service.worker_scratch_dir

        self.local_temp_dir = tempfile.mkdtemp()

        self.logger.info("local_temp_dir: %s" % self.local_temp_dir)
        self.logger.info("node: %s" % str(self.node.account))

    @cluster(num_nodes=1)
    def test_copy_to_dir_with_rename(self):
        # make dir structure locally
        make_dir_structure(self.local_temp_dir, DIR_STRUCTURE)
        dest = os.path.join(self.remote_scratch_dir, "renamed")
        self.node.account.copy_to(self.local_temp_dir, dest)

        # now validate the directory structure on the remote machine
        verify_dir_structure(dest, DIR_STRUCTURE, node=self.node)

    @cluster(num_nodes=1)
    def test_copy_to_dir_as_subtree(self):
        # copy directory "into" a directory; this should preserve the original directoryname
        make_dir_structure(self.local_temp_dir, DIR_STRUCTURE)
        self.node.account.copy_to(self.local_temp_dir, self.remote_scratch_dir)
        local_temp_dir_name = self.local_temp_dir
        if local_temp_dir_name.endswith(os.path.sep):
            local_temp_dir_name = local_temp_dir_name[:-len(os.path.sep)]

        verify_dir_structure(os.path.join(self.remote_scratch_dir, local_temp_dir_name), DIR_STRUCTURE)

    @cluster(num_nodes=1)
    def test_copy_from_dir_with_rename(self):
        # make dir structure remotely
        make_dir_structure(self.remote_scratch_dir, DIR_STRUCTURE, node=self.node)
        dest = os.path.join(self.local_temp_dir, "renamed")
        self.node.account.copy_from(self.remote_scratch_dir, dest)

        # now validate the directory structure locally
        verify_dir_structure(dest, DIR_STRUCTURE)

    @cluster(num_nodes=1)
    def test_copy_from_dir_as_subtree(self):
        # copy directory "into" a directory; this should preserve the original directoryname
        make_dir_structure(self.remote_scratch_dir, DIR_STRUCTURE, node=self.node)
        self.node.account.copy_from(self.remote_scratch_dir, self.local_temp_dir)

        verify_dir_structure(os.path.join(self.local_temp_dir, "scratch"), DIR_STRUCTURE)

    def teardown(self):
        # allow_fail in case scratch dir was not successfully created
        if os.path.exists(self.local_temp_dir):
            shutil.rmtree(self.local_temp_dir)


class CopyDirectTest(Test):

    def setup(self):
        self.service = GenericService(self.test_context, 2)
        self.src_node, self.dest_node = self.service.nodes
        self.remote_scratch_dir = self.service.worker_scratch_dir

        self.logger.info("src_node: %s" % str(self.src_node.account))
        self.logger.info("dest_node: %s" % str(self.dest_node.account))

    @cluster(num_nodes=2)
    def test_copy_file(self):
        """Verify that a file can be correctly copied directly between nodes.

        This should work with or without the recursive flag.
        """
        file_path = os.path.join(self.remote_scratch_dir, "myfile.txt")
        expected_contents = b"123"
        self.src_node.account.create_file(file_path, expected_contents)

        self.src_node.account.copy_between(file_path, file_path, self.dest_node)

        assert self.dest_node.account.isfile(file_path)
        with self.dest_node.account.open(file_path, "r") as f:
            contents = f.read()
            assert expected_contents == contents

    @cluster(num_nodes=2)
    def test_copy_directory(self):
        """Verify that a directory can be correctly copied directly between nodes.
        """

        make_dir_structure(self.remote_scratch_dir, DIR_STRUCTURE, node=self.src_node)
        self.src_node.account.copy_between(self.remote_scratch_dir, self.remote_scratch_dir, self.dest_node)
        verify_dir_structure(os.path.join(self.remote_scratch_dir, "scratch"), DIR_STRUCTURE, node=self.dest_node)


class TestClusterSpec(Test):
    @cluster(cluster_spec=ClusterSpec.simple_linux(2))
    def test_create_two_node_service(self):
        self.service = GenericService(self.test_context, 2)
        for node in self.service.nodes:
            node.account.ssh("echo hi")

    @cluster(cluster_spec=ClusterSpec.from_nodes(
        [
            NodeSpec(operating_system=WINDOWS),
            NodeSpec(operating_system=LINUX),
            NodeSpec()  # this one is also linux
        ]
    ))
    def three_nodes_test(self):
        self.service = GenericService(self.test_context, 3)
        for node in self.service.nodes:
            node.account.ssh("echo hi")


class RemoteAccountTest(Test):
    def __init__(self, test_context):
        super(RemoteAccountTest, self).__init__(test_context)
        self.account_service = RemoteAccountTestService(test_context)

    def setup(self):
        self.account_service.start()

    @cluster(num_nodes=1)
    def test_flaky(self):
        assert random.choice([True, False, False])

    @cluster(num_nodes=1)
    def test_ssh_capture_combine_stderr(self):
        """Test that ssh_capture correctly captures stderr and stdout from remote process.
        """
        node = self.account_service.nodes[0]

        # swap stdout and stderr in the echo process
        cmd = "for i in $(seq 1 5); do echo $i 3>&1 1>&2 2>&3; done"

        ssh_output = node.account.ssh_capture(cmd, combine_stderr=True)
        bad_ssh_output = node.account.ssh_capture(cmd, combine_stderr=False)  # Same command, but don't capture stderr

        lines = [int(line.strip()) for line in ssh_output]
        assert lines == [i for i in range(1, 6)]
        bad_lines = [int(line.strip()) for line in bad_ssh_output]
        assert bad_lines == []

    @cluster(num_nodes=1)
    def test_ssh_output_combine_stderr(self):
        """Test that ssh_output correctly captures stderr and stdout from remote process.
        """
        node = self.account_service.nodes[0]

        # swap stdout and stderr in the echo process
        cmd = "for i in $(seq 1 5); do echo $i 3>&1 1>&2 2>&3; done"

        ssh_output = node.account.ssh_output(cmd, combine_stderr=True)
        bad_ssh_output = node.account.ssh_output(cmd, combine_stderr=False)  # Same command, but don't capture stderr

        assert ssh_output == b"\n".join([str(i).encode('utf-8') for i in range(1, 6)]) + b"\n", ssh_output
        assert bad_ssh_output == b"", bad_ssh_output

    @cluster(num_nodes=1)
    def test_ssh_capture(self):
        """Test that ssh_capture correctly captures output from ssh subprocess.
        """
        node = self.account_service.nodes[0]
        cmd = "for i in $(seq 1 5); do echo $i; done"
        ssh_output = node.account.ssh_capture(cmd, combine_stderr=False)

        lines = [int(line.strip()) for line in ssh_output]
        assert lines == [i for i in range(1, 6)]

    @cluster(num_nodes=1)
    def test_ssh_output(self):
        """Test that ssh_output correctly captures output from ssh subprocess.
        """
        node = self.account_service.nodes[0]
        cmd = "for i in $(seq 1 5); do echo $i; done"
        ssh_output = node.account.ssh_output(cmd, combine_stderr=False)

        assert ssh_output == b"\n".join([str(i).encode('utf-8') for i in range(1, 6)]) + b"\n", ssh_output

    @cluster(num_nodes=1)
    def test_monitor_log(self):
        """Tests log monitoring by writing to a log in the background thread"""

        node = self.account_service.nodes[0]

        # Make sure we start the log with some data, including the value we're going to grep for
        self.account_service.write_to_log("foo\nbar\nbaz")

        # Background thread that simulates a process writing to the log
        self.wrote_log_line = False

        def background_logging_thread():
            # This needs to be large enough that we can verify we've actually
            # waited some time for the data to be written, but not too long that
            # the test takes a long time
            time.sleep(3)
            self.wrote_log_line = True
            self.account_service.write_to_log("foo\nbar\nbaz")

        with node.account.monitor_log(self.account_service.log_file) as monitor:
            logging_thread = Thread(target=background_logging_thread)
            logging_thread.start()
            monitor.wait_until('foo', timeout_sec=10, err_msg="Never saw expected log")
            assert self.wrote_log_line

        logging_thread.join(5.0)
        if logging_thread.isAlive():
            raise Exception("Timed out waiting for background thread.")

    @cluster(num_nodes=1)
    def test_monitor_log_exception(self):
        """Tests log monitoring correctly throws an exception when the regex was not found"""

        node = self.account_service.nodes[0]

        # Make sure we start the log with some data, including the value we're going to grep for
        self.account_service.write_to_log("foo\nbar\nbaz")

        timeout = 3
        try:
            with node.account.monitor_log(self.account_service.log_file) as monitor:
                start = time.time()
                monitor.wait_until('foo', timeout_sec=timeout, err_msg="Never saw expected log")
                assert False, "Log monitoring should have timed out and thrown an exception"
        except TimeoutError:
            # expected
            end = time.time()
            assert end - start > timeout, "Should have waited full timeout period while monitoring the log"

    @cluster(num_nodes=1)
    def test_kill_process(self):
        """Tests that kill_process correctly works"""

        def get_pids():
            pid_cmd = "ps ax | grep -i nc | grep -v grep | awk '{print $1}'"

            return list(node.account.ssh_capture(pid_cmd, callback=int))

        node = self.account_service.nodes[0]

        # Run TCP service using netcat
        node.account.ssh_capture("nohup nc -l -p 5000 > /dev/null 2>&1 &")

        wait_until(lambda: len(get_pids()) > 0, timeout_sec=10,
                   err_msg="Failed to start process within %d sec" % 10)

        # Kill service.
        node.account.kill_process("nc")

        wait_until(lambda: len(get_pids()) == 0, timeout_sec=10,
                   err_msg="Failed to kill process within %d sec" % 10)


class TestIterWrapper(Test):
    def setup(self):
        self.line_num = 6
        self.eps = 0.01

        self.service = GenericService(self.test_context, num_nodes=1)
        self.node = self.service.nodes[0]

        self.temp_file = "ducktape-test-" + str(random.randint(0, 100000))
        contents = ""
        for i in range(self.line_num):
            contents += "%d\n" % i

        self.node.account.create_file(self.temp_file, contents)

    def test_iter_wrapper(self):
        """Test has_next functionality on the returned iterable item."""
        output = self.node.account.ssh_capture("cat " + self.temp_file)
        for i in range(self.line_num):
            assert output.has_next()  # with timeout in case of hang
            assert output.next().strip() == str(i)
        start = time.time()
        assert output.has_next() is False
        stop = time.time()
        assert stop - start < self.eps, "has_next() should return immediately"

    def test_iter_wrapper_timeout(self):
        """Test has_next with timeout"""
        output = self.node.account.ssh_capture("tail -F " + self.temp_file)
        # allow command to be executed before we check output with timeout_sec = 0
        time.sleep(.5)
        for i in range(self.line_num):
            assert output.has_next(timeout_sec=0)
            assert output.next().strip() == str(i)

        timeout = .25
        start = time.time()
        # This check will last for the duration of the timeout because the the remote tail -F process
        # remains running, and the output stream is not closed.
        assert output.has_next(timeout_sec=timeout) is False
        stop = time.time()
        assert (stop - start >= timeout) and (stop - start) < timeout + self.eps, \
            "has_next() should return right after %s second" % str(timeout)

    def teardown(self):
        # tail -F call above will leave stray processes, so clean up
        cmd = "for p in $(ps ax | grep -v grep | grep \"%s\" | awk '{print $1}'); do kill $p; done" % self.temp_file
        self.node.account.ssh(cmd, allow_fail=True)

        self.node.account.ssh("rm -f " + self.temp_file, allow_fail=True)


class RemoteAccountCompressedTest(Test):
    def __init__(self, test_context):
        super(RemoteAccountCompressedTest, self).__init__(test_context)
        self.account_service = RemoteAccountTestService(test_context)
        self.test_context.session_context.compress = True
        self.tar_msg = False
        self.tar_error = False

    def setup(self):
        self.account_service.start()

    @cluster(num_nodes=1)
    def test_log_compression_with_non_existent_files(self):
        """Test that log compression with tar works even when a specific log file has not been generated
        (e.g. heap dump)
        """
        self.test_context.logger.addFilter(CompressionErrorFilter(self))
        self.copy_service_logs(None)

        if not self.tar_msg:
            raise Exception("Never saw attempt to compress log")
        if self.tar_error:
            raise Exception("Failure when compressing logs")


class CompressionErrorFilter(logging.Filter):

    def __init__(self, test):
        super(CompressionErrorFilter, self).__init__()
        self.test = test

    def filter(self, record):
        if 'tar czf' in record.msg:
            self.test.tar_msg = True
            if 'Error' in record.msg:
                self.test.tar_error = True
        return True



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster_spec import LINUX
from ducktape.cluster.finite_subcluster import FiniteSubcluster
from ducktape.cluster.node_container import InsufficientResourcesError, NodeNotPresentError
from ducktape.services.service import Service
import pickle
import pytest


class MockFiniteSubclusterNode:
    @property
    def operating_system(self):
        return LINUX


class CheckFiniteSubcluster(object):
    single_node_cluster_json = {"nodes": [{"hostname": "localhost"}]}

    def check_cluster_size(self):
        cluster = FiniteSubcluster([])
        assert len(cluster) == 0

        n = 10
        cluster = FiniteSubcluster([MockFiniteSubclusterNode() for _ in range(n)])
        assert len(cluster) == n

    def check_pickleable(self):
        cluster = FiniteSubcluster([MockFiniteSubclusterNode() for _ in range(10)])
        pickle.dumps(cluster)

    def check_allocate_free(self):
        n = 10
        cluster = FiniteSubcluster([MockFiniteSubclusterNode() for _ in range(n)])
        assert len(cluster) == n
        assert cluster.num_available_nodes() == n

        nodes = cluster.alloc(Service.setup_cluster_spec(num_nodes=1))
        assert len(nodes) == 1
        assert len(cluster) == n
        assert cluster.num_available_nodes() == n - 1

        nodes2 = cluster.alloc(Service.setup_cluster_spec(num_nodes=2))
        assert len(nodes2) == 2
        assert len(cluster) == n
        assert cluster.num_available_nodes() == n - 3

        cluster.free(nodes)
        assert cluster.num_available_nodes() == n - 2

        cluster.free(nodes2)
        assert cluster.num_available_nodes() == n

    def check_alloc_too_many(self):
        n = 10
        cluster = FiniteSubcluster([MockFiniteSubclusterNode() for _ in range(n)])
        with pytest.raises(InsufficientResourcesError):
            cluster.alloc(Service.setup_cluster_spec(num_nodes=(n + 1)))

    def check_free_too_many(self):
        n = 10
        cluster = FiniteSubcluster([MockFiniteSubclusterNode() for _ in range(n)])
        nodes = cluster.alloc(Service.setup_cluster_spec(num_nodes=n))
        with pytest.raises(NodeNotPresentError):
            nodes.append(MockFiniteSubclusterNode())
            cluster.free(nodes)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.mark.mark_expander import MarkedFunctionExpander
from ducktape.mark import env, is_env

import os


class CheckEnv(object):

    def check_does_not_raise_exception_when_key_not_exists(self):
        class C(object):
            @env(BLAH='8')
            def function(self):
                return 1

    def check_has_env_annotation(self):
        class C(object):
            @env(JAVA_HOME="blah")
            def function(self):
                return 1

        assert is_env(C.function)

    def check_is_ignored_if_env_not_correct(self):
        class C(object):
            @env(JAVA_HOME="blah")
            def function(self):
                return 1

        context_list = MarkedFunctionExpander(function=C.function, cls=C).expand()
        assert context_list[0].ignore

    def check_is_not_ignore_if_correct_env(self):
        os.environ['test_key'] = 'test'

        class C(object):
            @env(test_key='test')
            def function(self):
                return 1

        context_list = MarkedFunctionExpander(function=C.function, cls=C).expand()
        assert not context_list[0].ignore



# Copyright 2017 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json

from ducktape.cluster.node_container import NodeContainer

LINUX = "linux"

WINDOWS = "windows"

SUPPORTED_OS_TYPES = [LINUX, WINDOWS]


class NodeSpec(object):
    """
    The specification for a ducktape cluster node.

    :param operating_system:    The operating system of the node.
    """
    def __init__(self, operating_system=LINUX):
        self.operating_system = operating_system
        if self.operating_system not in SUPPORTED_OS_TYPES:
            raise RuntimeError("Unsupported os type %s" % self.operating_system)

    def __str__(self):
        dict = {
            "os": self.operating_system,
        }
        return json.dumps(dict, sort_keys=True)


class ClusterSpec(object):
    """
    The specification for a ducktape cluster.
    """
    nodes: NodeContainer = None

    @staticmethod
    def empty():
        return ClusterSpec([])

    @staticmethod
    def simple_linux(num_nodes):
        """
        Create a ClusterSpec containing some simple Linux nodes.
        """
        node_specs = [NodeSpec(LINUX)] * num_nodes
        return ClusterSpec(node_specs)

    @staticmethod
    def from_nodes(nodes):
        """
        Create a ClusterSpec describing a list of nodes.
        """
        return ClusterSpec(ClusterSpec([NodeSpec(node.operating_system) for node in nodes]))

    def __init__(self, nodes=None):
        """
        Initialize the ClusterSpec.

        :param nodes:           A collection of NodeSpecs, or None to create an empty cluster spec.
        """
        self.nodes = NodeContainer(nodes)

    def __len__(self):
        return self.size()

    def __iter__(self):
        return self.nodes.elements()

    def size(self):
        """Return the total size of this cluster spec, including all types of nodes."""
        return self.nodes.size()

    def add(self, other):
        """
        Add another ClusterSpec to this one.

        :param node_spec:       The other cluster spec.  This will not be modified.
        :return:                This ClusterSpec.
        """
        for node_spec in other.nodes:
            self.nodes.add_node(node_spec)
        return self

    def clone(self):
        """
        Returns a deep copy of this object.
        """
        return ClusterSpec(self.nodes.clone())

    def __str__(self):
        node_spec_to_num = {}
        for node_spec in self.nodes.elements():
            node_spec_str = str(node_spec)
            node_spec_to_num[node_spec_str] = node_spec_to_num.get(node_spec_str, 0) + 1
        rval = []
        for node_spec_str in sorted(node_spec_to_num.keys()):
            node_spec = json.loads(node_spec_str)
            node_spec["num_nodes"] = node_spec_to_num[node_spec_str]
            rval.append(node_spec)
        return json.dumps(rval, sort_keys=True)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ducktape.cluster.json import JsonCluster
from ducktape.cluster.node_container import InsufficientResourcesError
from ducktape.services.service import Service
import pickle
import pytest

from tests.runner.fake_remote_account import create_fake_remote_account


def create_json_cluster(*args, **kwargs):
    return JsonCluster(*args, make_remote_account_func=create_fake_remote_account, **kwargs)


class CheckJsonCluster(object):
    single_node_cluster_json = {
        "nodes": [
            {
                "ssh_config": {"host": "localhost"}
            }
        ]
    }

    def check_invalid_json(self):
        # Missing list of nodes
        with pytest.raises(ValueError):
            create_json_cluster({})

        # Missing hostname, which is required
        with pytest.raises(ValueError):
            create_json_cluster({"nodes": [{}]})

    @staticmethod
    def cluster_hostnames(nodes):
        return set([node.account.hostname for node in nodes])

    def check_cluster_size(self):
        cluster = create_json_cluster({"nodes": []})
        assert len(cluster) == 0

        n = 10
        cluster = create_json_cluster(
            {"nodes": [
                {"ssh_config": {"hostname": "localhost%d" % x}} for x in range(n)]})

        assert len(cluster) == n

    def check_pickleable(self):
        cluster = create_json_cluster(
            {"nodes": [
                {"ssh_config": {"host": "localhost1"}},
                {"ssh_config": {"host": "localhost2"}},
                {"ssh_config": {"host": "localhost3"}}]})

        pickle.dumps(cluster)

    def check_allocate_free(self):
        cluster = create_json_cluster(
            {"nodes": [
                {"ssh_config": {"host": "localhost1"}},
                {"ssh_config": {"host": "localhost2"}},
                {"ssh_config": {"host": "localhost3"}}]})

        assert len(cluster) == 3
        assert(cluster.num_available_nodes() == 3)

        nodes = cluster.alloc(Service.setup_cluster_spec(num_nodes=1))
        nodes_hostnames = self.cluster_hostnames(nodes)
        assert len(cluster) == 3
        assert(cluster.num_available_nodes() == 2)

        nodes2 = cluster.alloc(Service.setup_cluster_spec(num_nodes=2))
        nodes2_hostnames = self.cluster_hostnames(nodes2)
        assert len(cluster) == 3
        assert(cluster.num_available_nodes() == 0)

        assert(nodes_hostnames.isdisjoint(nodes2_hostnames))

        cluster.free(nodes)
        assert(cluster.num_available_nodes() == 1)

        cluster.free(nodes2)
        assert(cluster.num_available_nodes() == 3)

    def check_parsing(self):
        """ Checks that RemoteAccounts are generated correctly from input JSON"""

        node = create_json_cluster(
            {
                "nodes": [
                    {"ssh_config": {"host": "hostname"}}]}).alloc(Service.setup_cluster_spec(num_nodes=1))[0]

        assert node.account.hostname == "hostname"
        assert node.account.user is None

        ssh_config = {
            "host": "hostname",
            "user": "user",
            "hostname": "localhost",
            "port": 22
        }
        node = create_json_cluster({"nodes": [{"hostname": "hostname",
                                               "user": "user",
                                               "ssh_config": ssh_config}]}).alloc(
            Service.setup_cluster_spec(num_nodes=1))[0]

        assert node.account.hostname == "hostname"
        assert node.account.user == "user"

        # check ssh configs
        assert node.account.ssh_config.host == "hostname"
        assert node.account.ssh_config.user == "user"
        assert node.account.ssh_config.hostname == "localhost"
        assert node.account.ssh_config.port == 22

    def check_exhausts_supply(self):
        cluster = create_json_cluster(self.single_node_cluster_json)
        with pytest.raises(InsufficientResourcesError):
            cluster.alloc(Service.setup_cluster_spec(num_nodes=2))

    def check_node_names(self):
        cluster = create_json_cluster(
            {"nodes": [
                {"ssh_config": {"host": "localhost1"}},
                {"ssh_config": {"host": "localhost2"}},
                {"ssh_config": {"host": "localhost3"}}]})
        hosts = set(["localhost1", "localhost2", "localhost3"])
        nodes = cluster.alloc(cluster.available())
        assert hosts == set(node.name for node in nodes)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.services.service import Service
from tests.ducktape_mock import test_context, session_context
from ducktape.cluster.localhost import LocalhostCluster


class DummyService(Service):
    """Simple fake service class."""

    def __init__(self, context, num_nodes):
        super(DummyService, self).__init__(context, num_nodes)
        self.started_count = 0
        self.cleaned_count = 0
        self.stopped_count = 0

        self.started_kwargs = {}
        self.cleaned_kwargs = {}
        self.stopped_kwargs = {}

    def idx(self, node):
        return 1

    def start_node(self, node, **kwargs):
        super(DummyService, self).start_node(node, **kwargs)
        self.started_count += 1
        self.started_kwargs = kwargs

    def clean_node(self, node, **kwargs):
        super(DummyService, self).clean_node(node, **kwargs)
        self.cleaned_count += 1
        self.cleaned_kwargs = kwargs

    def stop_node(self, node, **kwargs):
        super(DummyService, self).stop_node(node, **kwargs)
        self.stopped_count += 1
        self.stopped_kwargs = kwargs


class DifferentDummyService(Service):
    """Another fake service class."""

    def __init__(self, context, num_nodes):
        super(DifferentDummyService, self).__init__(context, num_nodes)

    def idx(self, node):
        return 1


class CheckAllocateFree(object):

    def setup_method(self, _):
        self.cluster = LocalhostCluster()
        self.session_context = session_context()
        self.context = test_context(self.session_context, cluster=self.cluster)

    def check_allocate_free(self):
        """Check that allocating and freeing nodes works.

        This regression test catches the error with Service.free() introduced in v0.3.3 and fixed in v0.3.4
        """

        # Node allocation takes place during service instantiation
        initial_cluster_size = len(self.cluster)
        self.service = DummyService(self.context, 10)
        assert self.cluster.num_available_nodes() == initial_cluster_size - 10

        self.service.free()
        assert self.cluster.num_available_nodes() == initial_cluster_size

    def check_order(self):
        """Check expected behavior with service._order method"""
        self.dummy0 = DummyService(self.context, 4)
        self.diffDummy0 = DifferentDummyService(self.context, 100)
        self.dummy1 = DummyService(self.context, 1)
        self.diffDummy1 = DifferentDummyService(self.context, 2)
        self.diffDummy2 = DifferentDummyService(self.context, 5)

        assert self.dummy0._order == 0
        assert self.dummy1._order == 1
        assert self.diffDummy0._order == 0
        assert self.diffDummy1._order == 1
        assert self.diffDummy2._order == 2


class CheckStartStop(object):

    def setup_method(self, _):
        self.cluster = LocalhostCluster()
        self.session_context = session_context()
        self.context = test_context(self.session_context, cluster=self.cluster)

    def check_start_stop_clean(self):
        """
        Checks that start, stop, and clean invoke the expected per-node calls, and that start also runs stop and
        clean
        """
        service = DummyService(self.context, 2)

        service.start()
        assert service.started_count == 2
        assert service.stopped_count == 2
        assert service.cleaned_count == 2

        service.stop()
        assert service.stopped_count == 4

        service.start(clean=False)
        assert service.started_count == 4
        assert service.stopped_count == 6
        assert service.cleaned_count == 2

        service.stop()
        assert service.stopped_count == 8

        service.clean()
        assert service.cleaned_count == 4

    def check_kwargs_support(self):
        """Check that start, stop, and clean, and their per-node versions, can accept keyword arguments"""
        service = DummyService(self.context, 2)

        kwargs = {"foo": "bar"}
        service.start(**kwargs)
        assert service.started_kwargs == kwargs
        service.stop(**kwargs)
        assert service.stopped_kwargs == kwargs
        service.clean(**kwargs)
        assert service.cleaned_kwargs == kwargs



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
import copy
import logging
import multiprocessing
import os
import signal
import time
import traceback
import zmq

from ducktape.cluster.node_container import InsufficientResourcesError
from ducktape.tests.serde import SerDe
from ducktape.tests.test import TestContext
from ducktape.command_line.defaults import ConsoleDefaults
from ducktape.tests.runner_client import run_client
from ducktape.tests.result import TestResults
from ducktape.utils.terminal_size import get_terminal_size
from ducktape.tests.event import ClientEventFactory, EventResponseFactory
from ducktape.cluster.finite_subcluster import FiniteSubcluster
from ducktape.tests.scheduler import TestScheduler
from ducktape.tests.result import FAIL, TestResult
from ducktape.tests.reporter import SimpleFileSummaryReporter, HTMLSummaryReporter, JSONReporter
from ducktape.utils import persistence
from ducktape.errors import TimeoutError


class Receiver(object):
    def __init__(self, min_port, max_port):
        assert min_port <= max_port, "Expected min_port <= max_port, but instead: min_port: %s, max_port %s" % \
                                     (min_port, max_port)
        self.port = None
        self.min_port = min_port
        self.max_port = max_port

        self.serde = SerDe()

        self.zmq_context = zmq.Context()
        self.socket = self.zmq_context.socket(zmq.REP)

    def start(self):
        """Bind to a random port in the range [self.min_port, self.max_port], inclusive
        """
        # note: bind_to_random_port may retry the same port multiple times
        self.port = self.socket.bind_to_random_port(addr="tcp://*", min_port=self.min_port, max_port=self.max_port + 1,
                                                    max_tries=2 * (self.max_port + 1 - self.min_port))

    def recv(self, timeout=1800000):
        if timeout is None:
            # use default value of 1800000 or 30 minutes
            timeout = 1800000
        self.socket.RCVTIMEO = timeout
        try:
            message = self.socket.recv()
        except zmq.Again:
            raise TimeoutError("runner client unresponsive")
        return self.serde.deserialize(message)

    def send(self, event):
        self.socket.send(self.serde.serialize(event))

    def close(self):
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.close()


TestKey = namedtuple('TestKey', ['test_id', 'test_index'])


class TestRunner(object):

    # When set to True, the test runner will finish running/cleaning the current test, but it will not run any more
    stop_testing = False

    def __init__(self, cluster, session_context, session_logger, tests, deflake_num,
                 min_port=ConsoleDefaults.TEST_DRIVER_MIN_PORT,
                 max_port=ConsoleDefaults.TEST_DRIVER_MAX_PORT):

        # Set handler for SIGTERM (aka kill -15)
        # Note: it doesn't work to set a handler for SIGINT (Ctrl-C) in this parent process because the
        # handler is inherited by all forked child processes, and it prevents the default python behavior
        # of translating SIGINT into a KeyboardInterrupt exception
        signal.signal(signal.SIGTERM, self._propagate_sigterm)

        # session_logger, message logger,
        self.session_logger = session_logger
        self.cluster = cluster
        self.event_response = EventResponseFactory()
        self.hostname = "localhost"
        self.receiver = Receiver(min_port, max_port)

        self.deflake_num = deflake_num

        self.session_context = session_context
        self.max_parallel = session_context.max_parallel
        self.results = TestResults(self.session_context, self.cluster)

        self.exit_first = self.session_context.exit_first

        self.main_process_pid = os.getpid()
        self.scheduler = TestScheduler(tests, self.cluster)

        self.test_counter = 1
        self.total_tests = len(self.scheduler)
        # This immutable dict tracks test_id -> test_context
        self._test_context = persistence.make_dict(**{t.test_id: t for t in tests})
        self._test_cluster = {}  # Track subcluster assigned to a particular TestKey
        self._client_procs = {}  # track client processes running tests
        self.active_tests = {}
        self.finished_tests = {}
        self.test_schedule_log = []

    def _propagate_sigterm(self, signum, frame):
        """Handler SIGTERM and SIGINT by propagating SIGTERM to all client processes.

        Note that multiprocessing processes are in the same process group as the main process, so Ctrl-C will
        result in SIGINT being propagated to all client processes automatically. This may result in multiple SIGTERM
        signals getting sent to client processes in quick succession.

        However, it is possible that the main process (and not the process group) receives a SIGINT or SIGTERM
        directly. Propagating SIGTERM to client processes is necessary in this case.
        """
        if os.getpid() != self.main_process_pid:
            # since we're using the multiprocessing module to create client processes,
            # this signal handler is also attached client processes, so we only want to propagate TERM signals
            # if this process *is* the main runner server process
            return

        self.stop_testing = True
        for p in self._client_procs.values():

            # this handler should be a noop if we're in a client process, so it's an error if the current pid
            # is in self._client_procs
            assert p.pid != os.getpid(), "Signal handler should not reach this point in a client subprocess."
            if p.is_alive():
                os.kill(p.pid, signal.SIGTERM)

    def who_am_i(self):
        """Human-readable name helpful for logging."""
        return self.__class__.__name__

    @property
    def _ready_to_trigger_more_tests(self):
        """Should we pull another test from the scheduler?"""
        return not self.stop_testing and \
            len(self.active_tests) < self.max_parallel and \
            self.scheduler.peek() is not None

    @property
    def _expect_client_requests(self):
        return len(self.active_tests) > 0

    def _report_unschedulable(self, unschedulable, err_msg=None):
        if not unschedulable:
            return

        self._log(logging.ERROR,
                  f"There are {len(unschedulable)} tests which cannot be run due to insufficient cluster resources")
        for tc in unschedulable:
            if err_msg:
                msg = err_msg
            else:
                msg = f"Test {tc.test_id} requires more resources than are available in the whole cluster. " \
                      f"{self.cluster.all().nodes.attempt_remove_spec(tc.expected_cluster_spec)}"

            self._log(logging.ERROR, msg)

            result = TestResult(
                tc,
                self.test_counter,
                self.session_context,
                test_status=FAIL,
                summary=msg,
                start_time=time.time(),
                stop_time=time.time())
            self.results.append(result)
            result.report()

            self.test_counter += 1

    def _check_unschedulable(self):
        self._report_unschedulable(self.scheduler.filter_unschedulable_tests())

    def run_all_tests(self):
        self.receiver.start()
        self.results.start_time = time.time()

        # Report tests which cannot be run
        self._check_unschedulable()

        # Run the tests!
        self._log(logging.INFO, "starting test run with session id %s..." % self.session_context.session_id)
        self._log(logging.INFO, "running %d tests..." % len(self.scheduler))
        while self._ready_to_trigger_more_tests or self._expect_client_requests:
            try:
                while self._ready_to_trigger_more_tests:
                    next_test_context = self.scheduler.peek()
                    try:
                        self._preallocate_subcluster(next_test_context)
                    except InsufficientResourcesError:
                        # We were not able to allocate the subcluster for this test,
                        # this means not enough nodes passed health check.
                        # Don't mark this test as failed just yet, some other test might finish running and
                        # free up healthy nodes.
                        # However, if some nodes failed, cluster size changed too, so we need to check if
                        # there are any tests that can no longer be scheduled.
                        self._log(
                            logging.INFO,
                            f"Couldn't schedule test context {next_test_context} but we'll keep trying",
                            exc_info=True
                        )
                        self._check_unschedulable()
                    else:
                        # only remove the test from the scheduler once we've successfully allocated a subcluster for it
                        self.scheduler.remove(next_test_context)
                        self._run_single_test(next_test_context)

                if self._expect_client_requests:
                    try:
                        event = self.receiver.recv(timeout=self.session_context.test_runner_timeout)
                        self._handle(event)
                    except Exception as e:
                        err_str = "Exception receiving message: %s: %s" % (str(type(e)), str(e))
                        err_str += "\n" + traceback.format_exc(limit=16)
                        self._log(logging.ERROR, err_str)

                        # All processes are on the same machine, so treat communication failure as a fatal error
                        for proc in self._client_procs.values():
                            proc.terminate()
                        self._client_procs = {}
                        raise
            except KeyboardInterrupt:
                # If SIGINT is received, stop triggering new tests, and let the currently running tests finish
                self._log(logging.INFO,
                          "Received KeyboardInterrupt. Now waiting for currently running tests to finish...")
                self.stop_testing = True

        for proc in self._client_procs.values():
            proc.join()
        self.receiver.close()

        return self.results

    def _run_single_test(self, test_context):
        """Start a test runner client in a subprocess"""
        current_test_counter = self.test_counter
        self.test_counter += 1
        self._log(logging.INFO, "Triggering test %d of %d..." % (current_test_counter, self.total_tests))

        # Test is considered "active" as soon as we start it up in a subprocess
        test_key = TestKey(test_context.test_id, current_test_counter)
        self.active_tests[test_key] = True
        self.test_schedule_log.append(test_key)

        proc = multiprocessing.Process(
            target=run_client,
            args=[
                self.hostname,
                self.receiver.port,
                test_context.test_id,
                current_test_counter,
                TestContext.logger_name(test_context, current_test_counter),
                TestContext.results_dir(test_context, current_test_counter),
                self.session_context.debug,
                self.session_context.fail_bad_cluster_utilization,
                self.deflake_num
            ])

        self._client_procs[test_key] = proc
        proc.start()

    def _preallocate_subcluster(self, test_context):
        """Preallocate the subcluster which will be used to run the test.

        Side effect: store association between the test_id and the preallocated subcluster.

        :param test_context
        :return None
        """
        allocated = self.cluster.alloc(test_context.expected_cluster_spec)
        if len(self.cluster.available()) == 0 and self.max_parallel > 1 and not self._test_cluster:
            self._log(logging.WARNING,
                      "Test %s is using entire cluster. It's possible this test has no associated cluster metadata."
                      % test_context.test_id)

        self._test_cluster[TestKey(test_context.test_id, self.test_counter)] = FiniteSubcluster(allocated)

    def _handle(self, event):
        self._log(logging.DEBUG, str(event))

        if event["event_type"] == ClientEventFactory.READY:
            self._handle_ready(event)
        elif event["event_type"] in [ClientEventFactory.RUNNING,
                                     ClientEventFactory.SETTING_UP, ClientEventFactory.TEARING_DOWN]:
            self._handle_lifecycle(event)
        elif event["event_type"] == ClientEventFactory.FINISHED:
            self._handle_finished(event)
        elif event["event_type"] == ClientEventFactory.LOG:
            self._handle_log(event)
        else:
            raise RuntimeError("Received event with unknown event type: " + str(event))

    def _handle_ready(self, event):
        test_key = TestKey(event["test_id"], event["test_index"])
        test_context = self._test_context[event["test_id"]]
        subcluster = self._test_cluster[test_key]

        self.receiver.send(
            self.event_response.ready(event, self.session_context, test_context, subcluster))

    def _handle_log(self, event):
        self.receiver.send(self.event_response.log(event))
        self._log(event["log_level"], event["message"])

    def _handle_finished(self, event):
        test_key = TestKey(event["test_id"], event["test_index"])
        self.receiver.send(self.event_response.finished(event))

        result = event['result']
        if result.test_status == FAIL and self.exit_first:
            self.stop_testing = True

        # Transition this test from running to finished
        del self.active_tests[test_key]
        self.finished_tests[test_key] = event
        self.results.append(result)

        # Free nodes used by the test
        subcluster = self._test_cluster[test_key]
        self.cluster.free(subcluster.nodes)
        del self._test_cluster[test_key]

        # Join on the finished test process
        self._client_procs[test_key].join()

        # Report partial result summaries - it is helpful to have partial test reports available if the
        # ducktape process is killed with a SIGKILL partway through
        test_results = copy.copy(self.results)  # shallow copy
        reporters = [
            SimpleFileSummaryReporter(test_results),
            HTMLSummaryReporter(test_results),
            JSONReporter(test_results)
        ]
        for r in reporters:
            r.report()

        if self._should_print_separator:
            terminal_width, y = get_terminal_size()
            self._log(logging.INFO, "~" * int(2 * terminal_width / 3))

    @property
    def _should_print_separator(self):
        """The separator is the twiddle that goes in between tests on stdout.

        This only makes sense to print if tests are run sequentially (aka max_parallel = 1) since
        output from tests is interleaved otherwise.

        Also, we don't want to print the separator after the last test output has been received, so
        we check that there's more test output expected.
        """
        return self.session_context.max_parallel == 1 and \
            (self._expect_client_requests or self._ready_to_trigger_more_tests)

    def _handle_lifecycle(self, event):
        self.receiver.send(self.event_response._event_response(event))

    def _log(self, log_level, msg, *args, **kwargs):
        """Log to the service log of the current test."""
        self.session_logger.log(log_level, msg, *args, **kwargs)



# Copyright (c) 2009 Jason M Baker
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module contains persistent data structures from pysistence project, patched to work with
python 3.x.
"""


def not_implemented_method(*args, **kwargs):
    raise NotImplementedError('Cannot set values in a PDict')


class PDict(dict):
    """
    Persistent dict.
    """
    __setitem__ = not_implemented_method
    __delitem__ = not_implemented_method
    update = not_implemented_method
    clear = not_implemented_method
    pop = not_implemented_method
    popitem = not_implemented_method

    def _as_transient(self):
        return dict(self)

    def copy(self):
        return PDict(self)

    def without(self, *keys):
        new_dict = self._as_transient()
        for key in keys:
            del new_dict[key]
        return PDict(new_dict)

    def using(self, **kwargs):
        new_dict = self._as_transient()
        new_dict.update(kwargs)
        return PDict(new_dict)

    def __reduce__(self):
        # Pickling support in python 2.x and 3.x
        return make_dict, (self._as_transient(),)


make_dict = PDict



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster import Cluster
from ducktape.cluster.cluster_spec import ClusterSpec
from ducktape.cluster.node_container import NodeContainer


class FiniteSubcluster(Cluster):
    """This cluster class gives us a mechanism for allocating finite blocks of nodes from another cluster.
    """

    def __init__(self, nodes):
        super(FiniteSubcluster, self).__init__()
        self.nodes = nodes
        self._available_nodes = NodeContainer(nodes)
        self._in_use_nodes = NodeContainer()

    def do_alloc(self, cluster_spec):
        # there cannot be any bad nodes here,
        # since FiniteSubcluster operates on ClusterNode objects,
        # which are not checked for health by NodeContainer.remove_spec
        # however there could be an error, specifically if a test decides to alloc more nodes than are available
        # in a previous ducktape version this exception was raised by remove_spec
        # in this one, for consistency, we let the cluster itself deal with allocation errors
        good_nodes, bad_nodes = self._available_nodes.remove_spec(cluster_spec)
        self._in_use_nodes.add_nodes(good_nodes)
        return good_nodes

    def free_single(self, node):
        self._in_use_nodes.remove_node(node)
        self._available_nodes.add_node(node)

    def available(self):
        return ClusterSpec.from_nodes(self._available_nodes)

    def used(self):
        return ClusterSpec.from_nodes(self._in_use_nodes)



# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import


from ducktape.cluster.cluster_spec import ClusterSpec, WINDOWS
from ducktape.cluster.node_container import NodeContainer, InsufficientHealthyNodesError
from ducktape.command_line.defaults import ConsoleDefaults
from .cluster import Cluster, ClusterNode
from ducktape.cluster.linux_remoteaccount import LinuxRemoteAccount
from ducktape.cluster.windows_remoteaccount import WindowsRemoteAccount
from .remoteaccount import RemoteAccountSSHConfig

import json
import os
import traceback


def make_remote_account(ssh_config, *args, **kwargs):
    """Factory function for creating the correct RemoteAccount implementation."""

    if ssh_config.host and WINDOWS in ssh_config.host:
        return WindowsRemoteAccount(ssh_config, *args, **kwargs)
    else:
        return LinuxRemoteAccount(ssh_config, *args, **kwargs)


class JsonCluster(Cluster):
    """An implementation of Cluster that uses static settings specified in a cluster file or json-serializeable dict
    """

    def __init__(self, cluster_json=None, *args, make_remote_account_func=make_remote_account, **kwargs):
        """Initialize JsonCluster

        JsonCluster can be initialized from:
            - a json-serializeable dict
            - a "cluster_file" containing json

        :param cluster_json: a json-serializeable dict containing node information. If ``cluster_json`` is None,
               load from file
        :param cluster_file (optional): Overrides the default location of the json cluster file

        Example json with a local Vagrant cluster::

            {
              "nodes": [
                {
                  "externally_routable_ip": "192.168.50.151",

                  "ssh_config": {
                    "host": "worker1",
                    "hostname": "127.0.0.1",
                    "identityfile": "/path/to/private_key",
                    "password": null,
                    "port": 2222,
                    "user": "vagrant"
                  }
                },
                {
                  "externally_routable_ip": "192.168.50.151",

                  "ssh_config": {
                    "host": "worker2",
                    "hostname": "127.0.0.1",
                    "identityfile": "/path/to/private_key",
                    "password": null,
                    "port": 2223,
                    "user": "vagrant"
                  }
                }
              ]
            }

        """
        super(JsonCluster, self).__init__()
        self._available_accounts: NodeContainer = NodeContainer()
        self._bad_accounts: NodeContainer = NodeContainer()
        self._in_use_nodes: NodeContainer = NodeContainer()
        if cluster_json is None:
            # This is a directly instantiation of JsonCluster rather than from a subclass (e.g. VagrantCluster)
            cluster_file = kwargs.get("cluster_file")
            if cluster_file is None:
                cluster_file = ConsoleDefaults.CLUSTER_FILE
            cluster_json = json.load(open(os.path.abspath(cluster_file)))
        try:
            for ninfo in cluster_json["nodes"]:
                ssh_config_dict = ninfo.get("ssh_config")
                assert ssh_config_dict is not None, \
                    "Cluster json has a node without a ssh_config field: %s\n Cluster json: %s" % (ninfo, cluster_json)

                ssh_config = RemoteAccountSSHConfig(**ninfo.get("ssh_config", {}))
                remote_account = \
                    make_remote_account_func(ssh_config, ninfo.get("externally_routable_ip"),
                                             ssh_exception_checks=kwargs.get("ssh_exception_checks"))
                if remote_account.externally_routable_ip is None:
                    remote_account.externally_routable_ip = self._externally_routable_ip(remote_account)
                self._available_accounts.add_node(remote_account)
        except BaseException as e:
            msg = "JSON cluster definition invalid: %s: %s" % (e, traceback.format_exc(limit=16))
            raise ValueError(msg)
        self._id_supplier = 0

    def do_alloc(self, cluster_spec):
        try:
            good_nodes, bad_nodes = self._available_accounts.remove_spec(cluster_spec)
        except InsufficientHealthyNodesError as e:
            self._bad_accounts.add_nodes(e.bad_nodes)
            raise e

        # even in case of no exceptions, we can still run into bad nodes, so let's track them
        if bad_nodes:
            self._bad_accounts.add_nodes(bad_nodes)

        # now let's gather all the good ones and convert them into ClusterNode objects
        allocated_nodes = []
        for account in good_nodes:
            allocated_nodes.append(ClusterNode(account, slot_id=self._id_supplier))
            self._id_supplier += 1
        self._in_use_nodes.add_nodes(allocated_nodes)

        return allocated_nodes

    def free_single(self, node):
        self._in_use_nodes.remove_node(node)
        self._available_accounts.add_node(node.account)
        node.account.close()

    def _externally_routable_ip(self, account):
        return None

    def available(self):
        return ClusterSpec.from_nodes(self._available_accounts)

    def used(self):
        return ClusterSpec.from_nodes(self._in_use_nodes)



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import psutil
import shutil
import tempfile

from ducktape.tests.loggermaker import LoggerMaker, close_logger


class DummyFileLoggerMaker(LoggerMaker):
    def __init__(self, log_dir, n_handles):
        """Create a logger with n_handles file handles, with files in log_dir"""
        self.log_dir = log_dir
        self.n_handles = n_handles

    @property
    def logger_name(self):
        return "a.b.c"

    def configure_logger(self):
        for i in range(self.n_handles):
            fh = logging.FileHandler(os.path.join(self.log_dir, "log-" + str(i)))
            self._logger.addHandler(fh)


def open_files():
    # current process
    p = psutil.Process()
    return p.open_files()


class CheckLogger(object):
    def setup_method(self, _):
        self.temp_dir = tempfile.mkdtemp()

    def check_close_logger(self):
        """Check that calling close_logger properly cleans up resources."""
        initial_open_files = open_files()

        n_handles = 100
        log_maker = DummyFileLoggerMaker(self.temp_dir, n_handles)
        # accessing logger attribute lazily triggers configuration of logger
        the_logger = log_maker.logger

        assert len(open_files()) == len(initial_open_files) + n_handles
        close_logger(the_logger)
        assert len(open_files()) == len(initial_open_files)

    def teardown_method(self, _):
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)



from ducktape.cluster.cluster_spec import LINUX, WINDOWS
from ducktape.cluster.remoteaccount import RemoteAccount


class FakeRemoteAccount(RemoteAccount):

    def __init__(self, *args, is_available=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.os = LINUX
        self.is_available = is_available

    def available(self):
        return self.is_available

    def fetch_externally_routable_ip(self, *args, **kwargs):
        return 'fake ip'


class FakeWindowsRemoteAccount(FakeRemoteAccount):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.os = WINDOWS


def create_fake_remote_account(*args, **kwargs):
    return FakeRemoteAccount(*args, **kwargs)



from ducktape.mark.resource import cluster
from ducktape.services.service import Service
from ducktape.tests.test import Test


class FakeService(Service):
    pass


class BadActorTest(Test):

    @cluster(num_nodes=2)
    def test_too_many_nodes(self):
        """
        This test should fail to allocate and
        should be dealt with gracefully by the framework, marking it as failed
        and moving on.
        """
        FakeService(self.test_context, num_nodes=3)



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from json import JSONEncoder


class DucktapeJSONEncoder(JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "to_json"):
            # to_json may return a dict or array or other naturally json serializable object
            # this allows serialization to work correctly on nested items
            return obj.to_json()
        else:
            # Let the base class default method raise the TypeError
            return JSONEncoder.default(self, obj)
